@@ -20,9 +20,9 @@
 #define NONAMELESSUNION
 #include <windows.h>
 #include <stdio.h>
+#include <Rversion.h>
+#include <Startup.h>
 #include <stdlib.h>		/* for exit */
-#include "Rversion.h"
-#include "Startup.h"
 
 extern void cmdlineoptions(int, char **);
 extern int setupui(void);
