@@ -152,11 +152,11 @@ function(x, main = deparse(substitute(x)), sub = NULL, xlab = NULL,
         x <- data.matrix(x)
     dimd <- length(dx <- dim(x))
     if(dimd == 0 || any(dx == 0))
-        stop(paste(sQuote("x"), "must not have 0 dimensionality"))
+        stop("'x' must not have 0 dimensionality")
     if(length(list(...)))
-        warning(paste("extra argument(s)",
-                      paste(sQuote(names(list(...))), collapse = ", "),
-                      "disregarded."))
+        warning(gettextf("extra argument(s) %s will be disregarded",
+                         paste(sQuote(names(list(...))), collapse = ", ")),
+                domain = NA)
     ##-- Set up 'Ind' matrix : to contain indices and data
     Ind <- 1:dx[1]
     if(dimd > 1) {
@@ -186,7 +186,7 @@ function(x, main = deparse(substitute(x)), sub = NULL, xlab = NULL,
         if(is.logical(shade))
             shade <- c(2, 4)
         else if(any(shade <= 0) || length(shade) > 5)
-            stop("invalid shade specification")
+            stop("invalid 'shade' specification")
         extended <- TRUE
         shade <- sort(shade)
         breaks <- c(-Inf, - rev(shade), 0, shade, Inf)
@@ -229,7 +229,7 @@ function(x, main = deparse(substitute(x)), sub = NULL, xlab = NULL,
     }
     if (!is.null(sort)) {
         if(length(sort) != dimd)
-            stop("length(sort) doesn't conform to dim(x)")
+            stop("length of 'sort' does not conform to 'dim(x)'")
         ## Sort columns.
         Ind[,1:dimd] <- Ind[,sort]
         off <- off[sort]
@@ -337,7 +337,7 @@ function(x, main = deparse(substitute(x)), sub = NULL, xlab = NULL,
 
 mosaicplot.formula <-
 function(formula, data = NULL, ...,
-         main = deparse(substitute(data)), subset)
+         main = deparse(substitute(data)), subset, na.action = stats::na.omit)
 {
     main # force evaluation here
     m <- match.call(expand.dots = FALSE)
@@ -356,7 +356,7 @@ function(formula, data = NULL, ...,
         if(is.matrix(edata))
             m$data <- as.data.frame(data)
         m$main <- m$... <- NULL
-        m$na.action <- na.omit # tabulation would omit them
+        m$na.action <- na.action
         m[[1]] <- as.name("model.frame")
         mf <- eval(m, parent.frame())
         mosaicplot(table(mf), main = main, ...)
