@@ -14,7 +14,7 @@ ts <- function(data = NA, start = 1, end = numeric(0), frequency = 1,
                )
 {
     if(is.data.frame(data)) data <- data.matrix(data)
-#   if(!is.numeric(data)) stop("`data'  must be a numeric vector or matrix")
+#   if(!is.numeric(data)) stop("'data'  must be a numeric vector or matrix")
     if(is.matrix(data)) {
 	nseries <- ncol(data)
 	ndata <- nrow(data)
@@ -23,7 +23,7 @@ ts <- function(data = NA, start = 1, end = numeric(0), frequency = 1,
 	nseries <- 1
 	ndata <- length(data)
     }
-    if(ndata == 0) stop("ts object must have one or more observations")
+    if(ndata == 0) stop("'ts' object must have one or more observations")
 
     if(missing(frequency)) frequency <- 1/deltat
     else if(missing(deltat)) deltat <- 1/frequency
@@ -45,7 +45,7 @@ ts <- function(data = NA, start = 1, end = numeric(0), frequency = 1,
     else if(missing(start))
 	start <- end - (ndata - 1)/frequency
 
-    if(start > end) stop("start cannot be after end")
+    if(start > end) stop("'start' cannot be after 'end'")
     nobs <- floor((end - start) * frequency + 1.01)
 
     if(nobs != ndata)
@@ -86,9 +86,11 @@ hasTsp <- function(x)
     x
 }
 
-is.ts <- function (x) inherits(x, "ts") && length(x)
+is.ts <- function(x) inherits(x, "ts") && length(x)
 
-as.ts <- function (x)
+as.ts <- function(x, ...) UseMethod("as.ts")
+
+as.ts.default <- function(x, ...)
 {
     if (is.ts(x)) x
     else if(!is.null(xtsp <- tsp(x))) ts(x, xtsp[1], xtsp[2], xtsp[3])
@@ -111,7 +113,7 @@ as.ts <- function (x)
     tsps <- sapply(sers[tsser], tsp)
     freq <- mean(tsps[3,])
     if(max(abs(tsps[3,] - freq)) > getOption("ts.eps")) {
-        stop("Not all series have the same frequency")
+        stop("not all series have the same frequency")
     }
     if(union) {
         st <- min(tsps[1,])
@@ -120,7 +122,7 @@ as.ts <- function (x)
         st <- max(tsps[1,])
         en <- min(tsps[2,])
         if(st > en) {
-            warning("Non-intersecting series")
+            warning("non-intersecting series")
             return(NULL)
         }
     }
@@ -178,9 +180,9 @@ as.ts <- function (x)
     fixup <- if(is.null(nm)) seq(along = l) else nm == ""
     ## <NOTE>
     dep <- sapply(l[fixup], function(x) deparse(x)[1])
-    ## We could add support for `deparse.level' here by creating dep
+    ## We could add support for 'deparse.level' here by creating dep
     ## as in list.names() inside table().  But there is a catch: we
-    ## need deparse.level = 2 to get the `usual' deparsing when the
+    ## need deparse.level = 2 to get the 'usual' deparsing when the
     ## method is invoked by the generic ...
     ## </NOTE>
     if(is.null(nm)) return(dep)
@@ -225,7 +227,7 @@ ts.intersect <- function(..., dframe = FALSE)
 diff.ts <- function (x, lag = 1, differences = 1, ...)
 {
     if (lag < 1 | differences < 1)
-        stop("Bad value for lag or differences")
+        stop("bad value for 'lag' or 'differences'")
     if (lag * differences >= NROW(x)) return(x[0])
     ## <FIXME>
     ## lag() and its default method are defined in package ts, so we
@@ -374,7 +376,7 @@ print.ts <- function(x, calendar, ...)
                 attributes(x) <- NULL
                 names(x) <- tx
             }
-        } else { ##-- no `calendar' --
+        } else { ##-- no 'calendar' --
             header(x)
             attr(x, "class") <- attr(x, "tsp") <- attr(x, "na.action") <- NULL
         }
@@ -382,7 +384,7 @@ print.ts <- function(x, calendar, ...)
 	if(calendar && fr.x > 1) {
 	    tm <- time(x)
 	    t2 <- 1 + round(fr.x*((tm+0.001) %%1))
-	    p1 <- format(floor(tm))# yr
+	    p1 <- format(floor(zapsmall(tm)))# yr
 	    rownames(x) <-
 		if(fr.x == 12)
 		    paste(month.abb[t2], p1, sep=" ")
@@ -428,7 +430,7 @@ plot.ts <-
 		      cex=cex.main, font=font.main, col=col.main, ...)
 	    panel <- match.fun(panel)
 	    nser <- NCOL(x)
-	    if(nser > 10) stop("Can't plot more than 10 series as \"multiple\"")
+	    if(nser > 10) stop("cannot plot more than 10 series as \"multiple\"")
 	    if(is.null(main)) main <- xlabel
 	    nm <- colnames(x)
 	    if(is.null(nm)) nm <- paste("Series", 1:nser)
@@ -483,7 +485,7 @@ plot.ts <-
 	    if(missing(xy.labels)) xy.labels <- (n <= 150)
 	    if(!is.logical(xy.labels)) {
 		if(!is.character(xy.labels))
-		    stop("`xy.labels' must be logical or character")
+		    stop("'xy.labels' must be logical or character")
 		do.lab <- TRUE
 	    } else do.lab <- xy.labels
 
@@ -572,7 +574,7 @@ window.default <- function(x, start = NULL, end = NULL,
 
     if(!is.null(frequency) && !is.null(deltat) &&
        abs(frequency*deltat - 1) > ts.eps)
-        stop("frequency and deltat are both supplied and are inconsistent")
+        stop("'frequency' and 'deltat' are both supplied and are inconsistent")
     if (is.null(frequency) && is.null(deltat)) yfreq <- xfreq
     else if (is.null(deltat)) yfreq <- frequency
     else if (is.null(frequency)) yfreq <- 1/deltat
@@ -589,10 +591,10 @@ window.default <- function(x, start = NULL, end = NULL,
     else switch(length(start),
 		start,
 		start[1] + (start[2] - 1)/xfreq,
-		stop("Bad value for start"))
+		stop("bad value for 'start'"))
     if(start < xtsp[1] && !extend) {
 	start <- xtsp[1]
-	warning("start value not changed")
+	warning("'start' value not changed")
     }
 
     end <- if(is.null(end))
@@ -600,14 +602,14 @@ window.default <- function(x, start = NULL, end = NULL,
     else switch(length(end),
 		end,
 		end[1] + (end[2] - 1)/xfreq,
-		stop("Bad value for end"))
+		stop("bad value for 'end'"))
     if(end > xtsp[2] && !extend) {
 	end <- xtsp[2]
-	warning("end value not changed")
+	warning("'end' value not changed")
     }
 
     if(start > end)
-	stop("start cannot be after end")
+	stop("'start' cannot be after 'end'")
 
     if(!extend) {
         if(all(abs(start - xtime) > abs(start) * ts.eps))
@@ -629,8 +631,10 @@ window.default <- function(x, start = NULL, end = NULL,
         enoff <- floor((end - xtsp[2]) * xfreq + ts.eps)
         yend <- xtsp[2] + enoff/xfreq
         nold <- round(xfreq*(xtsp[2] - xtsp[1])) + 1
+        ## both start and end could be outside time base
+        i0 <- 1+max(0, stoff); i1 <- nold + min(0, enoff)
         i <- c(rep.int(nold+1, max(0, -stoff)),
-                   (1+max(0, stoff)):(nold + min(0, enoff)),
+                   if(i0 <= i1) i0:i1,
                    rep.int(nold+1, max(0, enoff)))
         y <- if(is.matrix(x)) rbind(x, NA)[i, , drop = FALSE] else c(x, NA)[i]
         attr(y, "tsp") <- c(ystart, yend, xfreq)
@@ -641,6 +645,40 @@ window.default <- function(x, start = NULL, end = NULL,
 
 window.ts <- function (x, ...) as.ts(window.default(x, ...))
 
+"window<-" <- function(x, ..., value) UseMethod("window<-")
+
+"window<-.ts" <- function(x, start, end, frequency, deltat, ..., value)
+{
+    xtsp <- tsp(x)
+    m <- match.call(expand.dots = FALSE)
+    m$value <- NULL
+    m$extend <- TRUE
+    m[[1]] <- as.name("window")
+    xx <- eval.parent(m)
+    xxtsp <- tsp(xx)
+    start <- xxtsp[1]; end <- xxtsp[2]
+    if(start > end) stop("'start' > 'end'")
+    if (start < xtsp[1] || end > xtsp[2]) {
+        warning("extending time series when replacing values", call. = FALSE)
+        x <- window(x, min(start, xtsp[1]), max(end, xtsp[2]), extend = TRUE)
+    }
+    xfreq <- xtsp[3]
+    xtimes <- round(xfreq*time(x))
+    xxtimes <- round(xfreq * time(xx))
+
+    ind <- match(xxtimes, xtimes)
+    if(any(is.na(ind))) stop("times to be replaced do not match")
+
+    len <- length(ind)
+    val_len <- length(value)
+    if(!val_len) stop("no replacement values supplied")
+    if(val_len > len) stop("too many replacement values supplied")
+    if(val_len > 1 && (len %% val_len))
+        stop("number of values supplied is not a sub-multiple of the number of values to be replaced")
+    if(NCOL(x) == 1) x[ind] <- value else x[ind, ] <- value
+    x
+}
+
 "[.ts" <- function (x, i, j, drop = TRUE) {
     y <- NextMethod("[")
     if (missing(i))
@@ -696,21 +734,21 @@ ts.plot <- function(..., gpars = list())
 arima.sim <- function(model, n, rand.gen = rnorm,
                       innov = rand.gen(n, ...), n.start = NA, ...)
 {
-    if(!is.list(model)) stop("`model' must be list")
+    if(!is.list(model)) stop("'model' must be list")
     p <- length(model$ar)
     if(p) {
         minroots <- min(Mod(polyroot(c(1, -model$ar))))
-        if(minroots <= 1) stop("ar part of model is not stationary")
+        if(minroots <= 1) stop("'ar' part of model is not stationary")
     }
     q <- length(model$ma)
     if(is.na(n.start)) n.start <- p + q +
         ifelse(p > 0, ceiling(6/log(minroots)), 0)
-    if(n.start < p + q) stop("burn-in must be as long as ar + ma")
+    if(n.start < p + q) stop("burn-in 'n.start' must be as long as 'ar + ma'")
     d <- 0
     if(!is.null(ord <- model$order)) {
-        if(length(ord) != 3) stop("`model$order' must be of length 3")
-        if(p != ord[1]) stop("inconsistent specification of ar order")
-        if(q != ord[3]) stop("inconsistent specification of ma order")
+        if(length(ord) != 3) stop("'model$order' must be of length 3")
+        if(p != ord[1]) stop("inconsistent specification of 'ar' order")
+        if(q != ord[3]) stop("inconsistent specification of 'ma' order")
         d <- ord[2]
         if(d != round(d) || d < 0)
             stop("number of differences must be a positive integer")
