@@ -15,8 +15,9 @@ Sweave <- function(file, driver=RweaveLatex(),
     drobj <- driver$setup(file=file, syntax=syntax, ...)
     on.exit(driver$finish(drobj, error=TRUE))
 
-    text <- readLines(file)
-
+    text <- SweaveReadFile(file, syntax)
+    syntax <- attr(text, "syntax")
+        
     mode <- "doc"
     chunknr <- 0
     chunk <- NULL
@@ -63,18 +64,11 @@ Sweave <- function(file, driver=RweaveLatex(),
             if(mode=="code" && any(grep(syntax$coderef, line))){
                 chunkref <- sub(syntax$coderef, "\\1", line)
                 if(!(chunkref %in% names(namedchunks)))
-                    warning(paste("Reference to unknown chunk",
-                                  chunkref))
+                    warning(gettextf("reference to unknown chunk %s", chunkref),
+                            domain = NA)
                 line <- namedchunks[[chunkref]]
             }
-            else if(mode=="doc" && any(grep(syntax$syntaxname, line))){
-                sname <- sub(syntax$syntaxname, "\\1", line)
-                syntax <- get(sname, mode = "list")
-                if(class(syntax) != "SweaveSyntax")
-                    stop(paste("Object '", sname,
-                               "'has not class SweaveSyntax"))
-                drobj$syntax <- syntax
-            }
+            
             if(is.null(chunk))
                 chunk <- line
             else
@@ -90,16 +84,103 @@ Sweave <- function(file, driver=RweaveLatex(),
     driver$finish(drobj)
 }
 
+SweaveReadFile <- function(file, syntax)
+{
+    ## file can be a vector to keep track of recursive calls to
+    ## SweaveReadFile.  In this case only the first element is
+    ## tried to read in, the rest are forbidden names for further
+    ## SweaveInput
+    f <- file[1]
+
+    bf <- basename(f)
+    df <- dirname(f)
+    if(!file.exists(f)){
+        f <- list.files(df, full.names=TRUE,
+                        pattern=paste(bf, syntax$extension, sep=""))
+
+        if(length(f)==0){
+            stop(gettextf("no Sweave file with name '%s' found", file[1]),
+                 domain = NA)
+        }
+        else if(length(f)>1){
+            stop(paste(gettextf("%s Sweave files for basename '%s' found:",
+                                length(f), file),
+                       paste("\n         ", f, collapse="")),
+                 domain = NA)
+        }
+    }
+
+    text <- readLines(f[1])
+    ## <FIXME>
+    ## This needs to be more refined eventually ...
+    if(any(is.na(nchar(text, "c")))) {
+        ## Ouch, invalid in the current locale.
+        ## (Can only happen in a MBCS locale.)
+        ## Try re-encoding from Latin1.
+        if(capabilities("iconv"))
+            text <- iconv(text, "latin1", "")
+        else
+            stop("Found invalid multi-byte character data.", "\n",
+                 "Cannot re-encode because iconv is not available.", "\n",
+                 "Try running R in a single-byte locale.")
+    }
+    ## </FIXME>
+
+    pos <- grep(syntax$syntaxname, text)
+    
+    if(length(pos)>1){
+        warning(gettextf("more than one syntax specification found, using the first one"), domain = NA)
+    }
+    if(length(pos)>0){    
+        sname <- sub(syntax$syntaxname, "\\1", text[pos[1]])
+        syntax <- get(sname, mode = "list")
+        if(class(syntax) != "SweaveSyntax")
+            stop(gettextf("object '%s' does not have class \"SweaveSyntax\"",
+                          sname), domain = NA)
+
+        text <- text[-pos]
+    }
+
+    if(!is.null(syntax$input)){
+        while(any(pos <- grep(syntax$input, text))){
+
+            pos <- pos[1]
+            ifile <- file.path(df, sub(syntax$input, "\\1", text[pos]))
+            if(any(ifile==file)){
+                stop(paste(gettextf("recursive Sweave input '%s' in stack",
+                                    ifile),
+                           paste("\n         ", 1:length(file), ": ",
+                                 rev(file), collapse="")),
+                 domain = NA)
+            }
+            itext <- SweaveReadFile(c(ifile, file), syntax)
+        
+            if(pos==1)
+                text <- c(itext, text[-pos])
+            else if(pos==length(text))
+                text <- c(text[-pos], itext)
+            else
+                text <- c(text[1:(pos-1)], itext, text[(pos+1):length(text)])
+        }
+    }
+
+    attr(text, "syntax") <- syntax
+    text
+}
+   
+    
+
 ###**********************************************************
 
 SweaveSyntaxNoweb <-
     list(doc = "^@",
          code = "^<<(.*)>>=.*",
          coderef = "^<<(.*)>>.*",
-         docopt = "\\\\SweaveOpts\\{([^\\}]*)\\}",
+         docopt = "^[[:space:]]*\\\\SweaveOpts\\{([^\\}]*)\\}",
          docexpr = "\\\\Sexpr\\{([^\\}]*)\\}",
          extension = "\\.[rsRS]?nw$",
-         syntaxname = "\\\\SweaveSyntax\\{([^\\}]*)\\}",
+         syntaxname = "^[[:space:]]*\\\\SweaveSyntax\\{([^\\}]*)\\}",
+         input = "^[[:space:]]*\\\\SweaveInput\\{([^\\}]*)\\}",
          trans = list(
              doc = "@",
              code = "<<\\1>>=",
@@ -107,7 +188,8 @@ SweaveSyntaxNoweb <-
              docopt = "\\\\SweaveOpts{\\1}",
              docexpr = "\\\\Sexpr{\\1}",
              extension = ".Snw",
-             syntaxname = "\\\\SweaveSyntax{SweaveSyntaxNoweb}")
+             syntaxname = "\\\\SweaveSyntax{SweaveSyntaxNoweb}",
+             input = "\\\\SweaveInput{\\1}")
          )
 
 class(SweaveSyntaxNoweb) <- "SweaveSyntax"
@@ -144,10 +226,10 @@ SweaveSyntConv <- function(file, syntax, output=NULL)
         syntax <- get(syntax)
 
     if(class(syntax) != "SweaveSyntax")
-        stop("Target syntax not of class `SweaveSyntax'.\n")
+        stop("target syntax not of class \"SweaveSyntax\"")
 
     if(is.null(syntax$trans))
-        stop("Target syntax contains no translation table.\n")
+        stop("target syntax contains no translation table")
 
     insynt <- SweaveGetSyntax(file)
     text = readLines(file)
@@ -187,7 +269,7 @@ SweaveParseOptions <- function(text, defaults=list(), check=NULL)
         return(defaults)
 
     if(any(sapply(x, length)!=2))
-        stop(paste("Parse error or empty option in\n", text))
+        stop(gettextf("parse error or empty option in\n%s", text), domain = NA)
 
     options <- defaults
 
@@ -260,9 +342,9 @@ RweaveLatexSetup <-
         if(.Platform$OS.type == "windows")
             styfile <- gsub("\\\\", "/", styfile)
         if(any(grep(" ", styfile)))
-            warning(paste("Path '", styfile, "' contains spaces,\n",
-                          "this may cause problems when running latex.",
-                          sep=""))
+            warning(gettextf("path to '%s' contains spaces,\n", styfile),
+                    gettext("this may cause problems when running LaTeX"),
+                    domain = NA)
     }
     else
         styfile <- "Sweave"
@@ -316,12 +398,12 @@ RweaveLatexRuncode <- function(object, chunk, options)
         chunkout <- object$output
 
     SweaveHooks(options, run=TRUE)
-    
+
     chunkexps <- try(parse(text=chunk), silent=TRUE)
     RweaveTryStop(chunkexps, options)
     openSinput <- FALSE
     openSchunk <- FALSE
-    
+
     if(length(chunkexps)==0)
         return(object)
 
@@ -362,12 +444,12 @@ RweaveLatexRuncode <- function(object, chunk, options)
         if(length(output)==1 & output[1]=="") output <- NULL
 
         RweaveTryStop(err, options)
-        
+
         if(object$debug)
             cat(paste(output, collapse="\n"))
 
         if(length(output)>0 & (options$results != "hide")){
-                                                        
+
             if(openSinput){
                 cat("\n\\end{Sinput}\n", file=chunkout, append=TRUE)
                 openSinput <- FALSE
@@ -383,9 +465,11 @@ RweaveLatexRuncode <- function(object, chunk, options)
             }
 
             output <- paste(output,collapse="\n")
-            if(options$strip.white){
+            if(options$strip.white %in% c("all", "true")){
                 output <- sub("^[[:space:]]*\n", "", output)
                 output <- sub("\n[[:space:]]*$", "", output)
+                if(options$strip.white=="all")
+                    output <- sub("\n[[:space:]]*\n", "\n", output)
             }
             cat(output, file=chunkout, append=TRUE)
             remove(output)
@@ -481,12 +565,13 @@ RweaveLatexWritedoc <- function(object, chunk)
 RweaveLatexFinish <- function(object, error=FALSE)
 {
     if(!object$quiet && !error)
-        cat(paste("\nYou can now run LaTeX on",
-                  summary(object$output)$description), "\n")
+        cat("\n",
+            gettextf("You can now run LaTeX on '%s'",
+                     summary(object$output)$description),
+            "\n", sep = "")
     close(object$output)
-    if(length(object$chunkout)>0){
+    if(length(object$chunkout) > 0)
         for(con in object$chunkout) close(con)
-    }
 }
 
 RweaveLatexOptions <- function(options)
@@ -499,7 +584,7 @@ RweaveLatexOptions <- function(options)
 
     NUMOPTS <- c("width", "height")
     NOLOGOPTS <- c(NUMOPTS, "results", "prefix.string",
-                   "engine", "label")
+                   "engine", "label", "strip.white")
 
     for(opt in names(options)){
         if(! (opt %in% NOLOGOPTS)){
@@ -508,16 +593,22 @@ RweaveLatexOptions <- function(options)
                 options[[opt]] <- c2l(options[[opt]])
             }
             if(is.na(options[[opt]]))
-                stop(paste("invalid value for", opt, ":", oldval))
+                stop(gettextf("invalid value for '%s' : %s", opt, oldval),
+                     domain = NA)
         }
         else if(opt %in% NUMOPTS){
             options[[opt]] <- as.numeric(options[[opt]])
         }
     }
 
+    options$results <- tolower(as.character(options$results))
     options$results <- match.arg(options$results,
                                  c("verbatim", "tex", "hide"))
 
+    options$strip.white <- tolower(as.character(options$strip.white))
+    options$strip.white <- match.arg(options$strip.white,
+                                     c("true", "false", "all"))
+
     options
 }
 
@@ -562,8 +653,8 @@ RweaveTryStop <- function(err, options){
         stop(msg, err, call.=FALSE)
     }
 }
-           
-        
+
+
 
 
 
@@ -610,7 +701,7 @@ RtangleSetup <- function(file, syntax,
 
     options <- list(split=split, prefix=prefix,
                     prefix.string=prefix.string,
-                    engine="R")
+                    engine="R", eval=TRUE)
 
     list(output=output, annotate=annotate, options=options,
          chunkout=list(), quiet=quiet, syntax=syntax)
@@ -642,7 +733,8 @@ RtangleRuncode <-  function(object, chunk, options)
     if(object$annotate){
         cat("###################################################\n",
             "### chunk number ", options$chunknr,
-            ": ", options$label, "\n",
+            ": ", options$label,
+            ifelse(options$eval, "", " eval=FALSE"), "\n",
             "###################################################\n",
             file=chunkout, append=TRUE, sep="")
     }
@@ -652,6 +744,9 @@ RtangleRuncode <-  function(object, chunk, options)
         cat("getOption(\"SweaveHooks\")[[\"", k, "\"]]()\n",
             file=chunkout, append=TRUE, sep="")
 
+    if(!options$eval)
+        chunk <- paste("##", chunk)
+    
     cat(chunk,"\n", file=chunkout, append=TRUE, sep="\n")
 
     if(is.null(options$label) & options$split)
@@ -666,7 +761,8 @@ RtangleWritedoc <- function(object, chunk)
     {
         opts <- sub(paste(".*", object$syntax$docopt, ".*", sep=""),
                     "\\1", chunk[pos[1]])
-        object$options <- SweaveParseOptions(opts, object$options)
+        object$options <- SweaveParseOptions(opts, object$options,
+                                             RweaveLatexOptions)
         chunk[pos[1]] <- sub(object$syntax$docopt, "", chunk[pos[1]])
     }
     return(object)
