@@ -18,6 +18,9 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/* <UTF8> char here is either ASCII or handled as a whole */
+
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -303,15 +306,8 @@ static void ctr_swapseg(SEGP seg)
 	/* ctr_segdir(): Determine the entry direction to the next cell */
 	/* and update the cell indices */
 
-#ifdef OLDcontour
-static double ctr_xtol;
-static double ctr_ytol;
-#define XMATCH(x0,x1) (fabs(x0-x1) < ctr_xtol)
-#define YMATCH(y0,y1) (fabs(y0-y1) < ctr_ytol)
-#else
 #define XMATCH(x0,x1) (fabs(x0-x1) == 0)
 #define YMATCH(y0,y1) (fabs(y0-y1) == 0)
-#endif
 
 static int ctr_segdir(double xend, double yend, double *x, double *y,
 		      int *i, int *j, int nx, int ny)
@@ -428,13 +424,13 @@ int findGapUp(double *xxx, double *yyy, int ns, double labelDistance,
 	/* Find a gap big enough for the label
 	   use several segments if necessary
 	*/
-	dX = xxx[jjj] - xxx[jjj - n - 1];
+	dX = xxx[jjj] - xxx[jjj - n - 1]; /* jjj - n - 1 == 0 */
 	dY = yyy[jjj] - yyy[jjj - n - 1];
 	dXC = GConvertXUnits(dX, USER, INCHES, dd);
 	dYC = GConvertYUnits(dY, USER, INCHES, dd);
 	distanceSum = hypot(dXC, dYC);
-	jjj = (jjj + 1);
-	n += 1;
+	jjj++;
+	n++;
     }
     if (distanceSum < labelDistance)
 	return 0;
@@ -454,78 +450,54 @@ int findGapDown(double *xxx, double *yyy, int ns, double labelDistance,
 	/* Find a gap big enough for the label
 	   use several segments if necessary
 	*/
-	dX = xxx[jjj] - xxx[jjj + n + 1];
+	dX = xxx[jjj] - xxx[jjj + n + 1]; /*jjj + n + 1 == ns -1 */
 	dY = yyy[jjj] - yyy[jjj + n + 1];
 	dXC = GConvertXUnits(dX, USER, INCHES, dd);
 	dYC = GConvertYUnits(dY, USER, INCHES, dd);
 	distanceSum = hypot(dXC, dYC);
 	jjj--;
-	n -= 1;
+	n++;
     }
     if (distanceSum < labelDistance)
 	return 0;
     else
-	return -n;
+	return n;
 }
 
-static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z, 
-		    double zc,
-		    SEXP labels, int cnum,
-		    Rboolean drawLabels, int method,
-		    Rboolean vectorFonts, int typeface, int fontindex,
-		    double atom, DevDesc *dd)
+/*
+ * Generate a list of segments for a single level
+ */
+static SEGP* contourLines(double *x, int nx, double *y, int ny,
+			 double *z, double zc, double atom)
 {
-/* draw a contour for one given contour level `zc' */
-
-/* maximal number of line segments of one contour segment: 
- * for preventing infinite loops -- shouldn't be needed --> warning */
-#define MAX_ns 25000
-
-    char *vmax;
-
     double f, xl, xh, yl, yh, zll, zhl, zlh, zhh, xx[4], yy[4];
-    double xend, yend;
-    int i, ii, j, jj, k, l, m, nacode, ns, ns2, dir;
-    SEGP seglist, seg, s, start, end;
-    double *xxx, *yyy;
-
-    double variance, dX, dY, deltaX, deltaY;
-    double dXC, dYC, deltaXC, deltaYC;
-    int range=0, indx=0, n; /* -Wall */
-    double lowestVariance;
-    double squareSum, sum;
-    int iii, jjj;
-    double distanceSum, labelDistance, avgGradient;
-    int zeroCount;
-    char buffer[255];
-    double avg;
-    int result;
-    double ux, uy, vx, vy;
-    double xStart, yStart;
-    double dx, dy, dxy;
-    double labelHeight;
-    SEXP label1 = PROTECT(allocVector(REALSXP, 8));
-    SEXP label2;
-    SEXP lab;
-    Rboolean gotLabel = FALSE;
-    Rboolean ddl;/* Don't draw label -- currently unused, i.e. always FALSE*/
-
-#ifdef DEBUG_contour
-    Rprintf("contour(lev = %g):\n", zc);
-#endif
+    int i, j, k, l, m, nacode;
+    SEGP seglist;
+    SEGP *segmentDB;
+    /* Initialize the segment data base */
+    /* Note we must be careful about resetting */
+    /* the top of the stack, otherwise we run out of */
+    /* memory after a sequence of displaylist replays */
+    /*
+     * This reset is done out in GEcontourLines
+     */
+    segmentDB = (SEGP*)R_alloc(nx*ny, sizeof(SEGP));
+    for (i = 0; i < nx; i++)
+	for (j = 0; j < ny; j++)
+	    segmentDB[i + j * nx] = NULL;
     for (i = 0; i < nx - 1; i++) {
-	xl = REAL(x)[i];
-	xh = REAL(x)[i + 1];
+	xl = x[i];
+	xh = x[i + 1];
 	for (j = 0; j < ny - 1; j++) {
-	    yl = REAL(y)[j];
-	    yh = REAL(y)[j + 1];
+	    yl = y[j];
+	    yh = y[j + 1];
 	    k = i + j * nx;
-	    zll = REAL(z)[k];
-	    zhl = REAL(z)[k + 1];
-	    zlh = REAL(z)[k + nx];
-	    zhh = REAL(z)[k + nx + 1];
+	    zll = z[k];
+	    zhl = z[k + 1];
+	    zlh = z[k + nx];
+	    zhh = z[k + nx + 1];
 
-	    /* If the value at a corner is exactly equal to a contour level, 
+	    /* If the value at a corner is exactly equal to a contour level,
 	     * change that value by a tiny amount */
 
 	    if (zll == zc) zll += atom;
@@ -662,9 +634,336 @@ static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
 		}
 		else error("k != 2 or 4");
 	    }
-	    ctr_SegDB[i + j * nx] = seglist;
+	    segmentDB[i + j * nx] = seglist;
 	}
     }
+    return segmentDB;
+}
+
+/* maximal number of line segments of one contour segment:
+ * for preventing infinite loops -- shouldn't be needed --> warning */
+#define MAX_ns 25000
+
+#define CONTOUR_LIST_STEP 100
+#define CONTOUR_LIST_LEVEL 0
+#define CONTOUR_LIST_X 1
+#define CONTOUR_LIST_Y 2
+
+static SEXP growList(SEXP oldlist) {
+    int i, len;
+    SEXP templist;
+    len = LENGTH(oldlist);
+    templist = PROTECT(allocVector(VECSXP, len + CONTOUR_LIST_STEP));
+    for (i=0; i<len; i++)
+	SET_VECTOR_ELT(templist, i, VECTOR_ELT(oldlist, i));
+    UNPROTECT(1);
+    return templist;
+}
+
+/*
+ * Store the list of segments for a single level in the SEXP
+ * list that will be returned to the user
+ */
+int addContourLines(double *x, int nx, double *y, int ny,
+		     double *z, double zc, double atom,
+		     SEGP* segmentDB, int nlines, SEXP container)
+{
+    double xend, yend;
+    int i, ii, j, jj, ns, ns2, dir, nc;
+    SEGP seglist, seg, s, start, end;
+    SEXP ctr, level, xsxp, ysxp, names;
+    /* Begin following contours. */
+    /* 1. Grab a segment */
+    /* 2. Follow its tail */
+    /* 3. Follow its head */
+    /* 4. Save the contour */
+    for (i = 0; i < nx - 1; i++)
+	for (j = 0; j < ny - 1; j++) {
+	    while ((seglist = segmentDB[i + j * nx])) {
+		ii = i; jj = j;
+		start = end = seglist;
+		segmentDB[i + j * nx] = seglist->next;
+		xend = seglist->x1;
+		yend = seglist->y1;
+		while ((dir = ctr_segdir(xend, yend, x, y,
+					 &ii, &jj, nx, ny))) {
+		    segmentDB[ii + jj * nx]
+			= ctr_segupdate(xend, yend, dir, TRUE,/* = tail */
+					segmentDB[ii + jj * nx], &seg);
+		    if (!seg) break;
+		    end->next = seg;
+		    end = seg;
+		    xend = end->x1;
+		    yend = end->y1;
+		}
+		end->next = NULL; /* <<< new for 1.2.3 */
+		ii = i; jj = j;
+		xend = seglist->x0;
+		yend = seglist->y0;
+		while ((dir = ctr_segdir(xend, yend, x, y,
+					 &ii, &jj, nx, ny))) {
+		    segmentDB[ii + jj * nx]
+			= ctr_segupdate(xend, yend, dir, FALSE,/* ie. head */
+					segmentDB[ii+jj*nx], &seg);
+		    if (!seg) break;
+		    seg->next = start;
+		    start = seg;
+		    xend = start->x0;
+		    yend = start->y0;
+		}
+
+		/* ns := #{segments of polyline} -- need to allocate */
+		s = start;
+		ns = 0;
+		/* MAX_ns: prevent inf.loop (shouldn't be needed) */
+		while (s && ns < MAX_ns) {
+		    ns++;
+		    s = s->next;
+		}
+		if(ns == MAX_ns)
+		    warning(_("contour(): circular/long seglist -- bug.report()!"));
+
+		/* countour midpoint : use for labelling sometime (not yet!) */
+		if (ns > 3) ns2 = ns/2; else ns2 = -1;
+
+		/*
+		 * "write" the contour locations into the list of contours
+		 */
+		ctr = PROTECT(allocVector(VECSXP, 3));
+		level = PROTECT(allocVector(REALSXP, 1));
+		xsxp = PROTECT(allocVector(REALSXP, ns + 1));
+		ysxp = PROTECT(allocVector(REALSXP, ns + 1));
+		REAL(level)[0] = zc;
+		SET_VECTOR_ELT(ctr, CONTOUR_LIST_LEVEL, level);
+		s = start;
+		REAL(xsxp)[0] = s->x0;
+		REAL(ysxp)[0] = s->y0;
+		ns = 1;
+		while (s->next && ns < MAX_ns) {
+		    s = s->next;
+		    REAL(xsxp)[ns] = s->x0;
+		    REAL(ysxp)[ns++] = s->y0;
+		}
+		REAL(xsxp)[ns] = s->x1;
+		REAL(ysxp)[ns] = s->y1;
+		SET_VECTOR_ELT(ctr, CONTOUR_LIST_X, xsxp);
+		SET_VECTOR_ELT(ctr, CONTOUR_LIST_Y, ysxp);
+		/*
+		 * Set the names attribute for the contour
+		 * So that users can extract components using
+		 * meaningful names
+		 */
+		PROTECT(names = allocVector(STRSXP, 3));
+		SET_STRING_ELT(names, 0, mkChar("level"));
+		SET_STRING_ELT(names, 1, mkChar("x"));
+		SET_STRING_ELT(names, 2, mkChar("y"));
+		setAttrib(ctr, R_NamesSymbol, names);
+		/*
+		 * We're about to add another line to the list ...
+		 */
+		nlines += 1;
+		nc = LENGTH(VECTOR_ELT(container, 0));
+		if (nlines == nc)
+		    /* Where does this get UNPROTECTed? */
+		    SET_VECTOR_ELT(container, 0,
+				   growList(VECTOR_ELT(container, 0)));
+		SET_VECTOR_ELT(VECTOR_ELT(container, 0), nlines - 1, ctr);
+		UNPROTECT(5);
+	    }
+	}
+    return nlines;
+}
+
+/*
+ * Given nx x values, ny y values, nx*ny z values,
+ * and nl cut-values in z ...
+ * ... produce a list of contour lines:
+ *   list of sub-lists
+ *     sub-list = x vector, y vector, and cut-value.
+ */
+SEXP GEcontourLines(double *x, int nx, double *y, int ny,
+		    double *z, double *levels, int nl,
+		    GEDevDesc *dd)
+{
+    char *vmax;
+    int i, nlines, len;
+    double atom, zmin, zmax;
+    SEGP* segmentDB;
+    SEXP container, mainlist, templist;
+    /*
+     * "tie-breaker" values
+     */
+    zmin = DBL_MAX;
+    zmax = DBL_MIN;
+    for (i = 0; i < nx * ny; i++)
+	if (R_FINITE(z[i])) {
+	    if (zmax < z[i]) zmax =  z[i];
+	    if (zmin > z[i]) zmin =  z[i];
+	}
+
+    if (zmin >= zmax) {
+	if (zmin == zmax)
+	    warning(_("all z values are equal"));
+	else
+	    warning(_("all z values are NA"));
+	return R_NilValue;
+    }
+    /* change to 1e-3, reconsidered because of PR#897
+     * but 1e-7, and even  2*DBL_EPSILON do not prevent inf.loop in contour().
+     * maybe something like   16 * DBL_EPSILON * (..).
+     * see also MAX_ns above */
+    atom = 1e-3 * (zmax - zmin);
+    /*
+     * Create a "container" which is a list with only 1 element.
+     * The element is the list of lines that will be built up.
+     * I create the container because this allows me to PROTECT
+     * the container once here and then UNPROTECT it at the end of
+     * this function and, as long as I always work with
+     * VECTOR_ELT(container, 0) and SET_VECTOR_ELT(container, 0)
+     * in functions called from here, I don't need to worry about
+     * protectin the list that I am building up.
+     * Why bother?  Because the list I am building can potentially
+     * grow and it's awkward to get the PROTECTs/UNPROTECTs right
+     * when you're in a loop and growing a list.
+     */
+    container = PROTECT(allocVector(VECSXP, 1));
+    /*
+     * Create "large" list (will trim excess at the end if necesary)
+     */
+    SET_VECTOR_ELT(container, 0, allocVector(VECSXP, CONTOUR_LIST_STEP));
+    nlines = 0;
+    /*
+     * Add lines for each contour level
+     */
+    for (i = 0; i < nl; i++) {
+	/*
+	 * The vmaxget/set is to manage the memory that gets
+	 * R_alloc'ed in the creation of the segmentDB structure
+	 */
+	vmax = vmaxget();
+	/*
+	 * Generate a segment database
+	 */
+	segmentDB = contourLines(x, nx, y, ny, z, levels[i], atom);
+	/*
+	 * Add lines to the list based on the segment database
+	 */
+	nlines = addContourLines(x, nx, y, ny, z, levels[i],
+				 atom, segmentDB, nlines,
+				 container);
+	vmaxset(vmax);
+    }
+    /*
+     * Trim the list of lines to the appropriate length.
+     */
+    len = LENGTH(VECTOR_ELT(container, 0));
+    if (nlines < len) {
+	mainlist = VECTOR_ELT(container, 0);
+	templist = PROTECT(allocVector(VECSXP, nlines));
+	for (i=0; i<nlines; i++)
+	    SET_VECTOR_ELT(templist, i, VECTOR_ELT(mainlist, i));
+	mainlist = templist;
+	UNPROTECT(1);  /* UNPROTECT templist */
+    } else
+	mainlist = VECTOR_ELT(container, 0);
+    UNPROTECT(1);  /* UNPROTECT container */
+    return mainlist;
+}
+
+SEXP GEdrawContourLines();
+
+SEXP do_contourLines(SEXP call, SEXP op, SEXP args, SEXP env)
+{
+    SEXP oargs, c, x, y, z;
+    int nx, ny, nc;
+    GEDevDesc *dd = GEcurrentDevice();
+
+    oargs = args;
+
+    x = CAR(args);
+    internalTypeCheck(call, x, REALSXP);
+    nx = LENGTH(x);
+    args = CDR(args);
+
+    y = CAR(args);
+    internalTypeCheck(call, y, REALSXP);
+    ny = LENGTH(y);
+    args = CDR(args);
+
+    z = CAR(args);
+    internalTypeCheck(call, z, REALSXP);
+    args = CDR(args);
+
+    /* levels */
+    c = CAR(args);
+    internalTypeCheck(call, c, REALSXP);
+    nc = LENGTH(c);
+    args = CDR(args);
+
+    return GEcontourLines(REAL(x), nx, REAL(y), ny, REAL(z), REAL(c), nc, dd);
+}
+
+/*
+ * The *base graphics* function contour() and the *general base*
+ * function contourLines() use the same code to generate contour lines
+ * (i.e., the function contourLines())
+ *
+ * I had a look at extracting the code that draws the labels
+ * into a *general base* function
+ * (e.g., into some sort of labelLines() function),
+ * but the code is too base-graphics-specific (e.g., one of the
+ * labelling methods seeks the location closest to the edge of the
+ * plotting region) so I've left it alone for now.
+ *
+ * This does mean that the contourLines() function is part of the
+ * graphics engine, but the contour() function is part of the
+ * base graphics system.
+ */
+
+static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
+		    double zc,
+		    SEXP labels, int cnum,
+		    Rboolean drawLabels, int method,
+		    Rboolean vectorFonts, int typeface, int fontindex,
+		    double atom, DevDesc *dd)
+{
+/* draw a contour for one given contour level 'zc' */
+
+    char *vmax;
+
+    double xend, yend;
+    int i, ii, j, jj, ns, ns2, dir;
+    SEGP seglist, seg, s, start, end;
+    double *xxx, *yyy;
+
+    double variance, dX, dY, deltaX, deltaY;
+    double dXC, dYC, deltaXC, deltaYC;
+    int range=0, indx=0, n; /* -Wall */
+    double lowestVariance;
+    double squareSum, sum;
+    int iii, jjj;
+    double distanceSum, labelDistance, avgGradient;
+    int zeroCount;
+    char buffer[255];
+    double avg;
+    int result;
+    double ux, uy, vx, vy;
+    double xStart, yStart;
+    double dx, dy, dxy;
+    double labelHeight;
+    SEXP label1 = PROTECT(allocVector(REALSXP, 8));
+    SEXP label2;
+    SEXP lab;
+    Rboolean gotLabel = FALSE;
+    Rboolean ddl;/* Don't draw label -- currently unused, i.e. always FALSE*/
+
+#ifdef DEBUG_contour
+    Rprintf("contour(lev = %g):\n", zc);
+#endif
+
+    vmax = vmaxget();
+    ctr_SegDB = contourLines(REAL(x), nx, REAL(y), ny, REAL(z), zc, atom);
+    /* we need to keep ctr_SegDB available, so vmaxset(vmax); was wrong */
 
     /* The segment database is now assembled. */
     /* Begin following contours. */
@@ -717,7 +1016,7 @@ static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
 		s = s->next;
 	    }
 	    if(ns == MAX_ns)
-		warning("contour(): circular/long seglist -- bug.report()!");
+		warning(_("contour(): circular/long seglist -- bug.report()!"));
 
 	    /* countour midpoint : use for labelling sometime (not yet!) */
 	    if (ns > 3) ns2 = ns/2; else ns2 = -1;
@@ -740,13 +1039,13 @@ static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
 #ifdef DEBUG_contour
 	    Rprintf("  [%2d,%2d]: (x,y)[1:%d] = ", i,j, ns);
 	    if(ns >= 5)
-		Rprintf(" (%g,%g), (%g,%g), ..., (%g,%g)\n", 
+		Rprintf(" (%g,%g), (%g,%g), ..., (%g,%g)\n",
 			xxx[0],yyy[0], xxx[1],yyy[1], xxx[ns-1],yyy[ns-1]);
 	    else
 		for(iii = 0; iii < ns; iii++)
 		    Rprintf(" (%g,%g)%s", xxx[iii],yyy[iii],
 			    (iii < ns-1) ? "," : "\n");
-#endif	    
+#endif
 
 	    GMode(1, dd);
 
@@ -770,10 +1069,10 @@ static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
 
 		if (vectorFonts) {
 		    /* 1, 1 => sans serif, basic font */
-		    labelDistance = 
+		    labelDistance =
 			GVStrWidth((unsigned char *)buffer, typeface, fontindex,
 				   INCHES, dd);
-		    labelHeight = 
+		    labelHeight =
 			GVStrHeight((unsigned char *)buffer, typeface, fontindex,
 				    INCHES, dd);
 		}
@@ -1052,17 +1351,18 @@ static void contour(SEXP x, int nx, SEXP y, int ny, SEXP z,
 	    vmaxset(vmax);
 	} /* while */
       } /* for(i .. )  for(j ..) */
+    vmaxset(vmax); /* now we are done with ctr_SegDB */
     UNPROTECT_PTR(label1); /* pwwwargh! This is messy, but last thing
 			      protected is likely labelList, and that needs
 			      to be preserved across calls */
 }
 
-/* contour(x, y, z, levels, labels, labcex, drawlabels, 
+/* contour(x, y, z, levels, labels, labcex, drawlabels,
  *         method, vfont, col = col, lty = lty, lwd = lwd)
  */
 SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
 {
-    SEXP oargs, c, x, y, z, vfont, col, lty, lwd, labels;
+    SEXP oargs, c, x, y, z, vfont, col, rawcol, lty, lwd, labels;
     int i, j, nx, ny, nc, ncol, nlty, nlwd;
     int ltysave, colsave, lwdsave;
     double cexsave;
@@ -1075,11 +1375,12 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
     int fontindex = 0;
     double labcex;
     DevDesc *dd = CurrentDevice();
+    SEXP result = R_NilValue;
 
     GCheckState(dd);
 
     if (length(args) < 4)
-	errorcall(call, "too few arguments");
+	errorcall(call, _("too few arguments"));
 
     oargs = args;
 
@@ -1116,7 +1417,7 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
 
     method = asInteger(CAR(args)); args = CDR(args);
     if (method < 1 || method > 3)
-	errorcall(call, "invalid value for \"method\"");
+	errorcall(call, _("invalid value for 'method'"));
 
     PROTECT(vfont = FixupVFont(CAR(args)));
     if (!isNull(vfont)) {
@@ -1126,7 +1427,8 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
     }
     args = CDR(args);
 
-    PROTECT(col = FixupCol(CAR(args), NA_INTEGER));
+    rawcol = CAR(args);
+    PROTECT(col = FixupCol(rawcol, R_TRANWHITE));
     ncol = length(col);
     args = CDR(args);
 
@@ -1139,31 +1441,31 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
     args = CDR(args);
 
     if (nx < 2 || ny < 2)
-	errorcall(call, "insufficient x or y values");
+	errorcall(call, _("insufficient 'x' or 'y' values"));
 
     if (nrows(z) != nx || ncols(z) != ny)
-	errorcall(call, "dimension mismatch");
+	errorcall(call, _("dimension mismatch"));
 
     if (nc < 1)
-	errorcall(call, "no contour values");
+	errorcall(call, _("no contour values"));
 
     for (i = 0; i < nx; i++) {
 	if (!R_FINITE(REAL(x)[i]))
-	    errorcall(call, "missing x values");
+	    errorcall(call, _("missing 'x' values"));
 	if (i > 0 && REAL(x)[i] < REAL(x)[i - 1])
-	    errorcall(call, "increasing x values expected");
+	    errorcall(call, _("increasing 'x' values expected"));
     }
 
     for (i = 0; i < ny; i++) {
 	if (!R_FINITE(REAL(y)[i]))
-	    errorcall(call, "missing y values");
+	    errorcall(call, _("missing 'y' values"));
 	if (i > 0 && REAL(y)[i] < REAL(y)[i - 1])
-	    errorcall(call, "increasing y values expected");
+	    errorcall(call, _("increasing 'y' values expected"));
     }
 
     for (i = 0; i < nc; i++)
 	if (!R_FINITE(REAL(c)[i]))
-	    errorcall(call, "illegal NA contour values");
+	    errorcall(call, _("invalid NA contour values"));
 
     zmin = DBL_MAX;
     zmax = DBL_MIN;
@@ -1175,9 +1477,9 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
 
     if (zmin >= zmax) {
 	if (zmin == zmax)
-	    warning("all z values are equal");
+	    warning(_("all z values are equal"));
 	else
-	    warning("all z values are NA");
+	    warning(_("all z values are NA"));
 	return R_NilValue;
     }
 
@@ -1207,17 +1509,21 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
     lwdsave = Rf_gpptr(dd)->lwd;
     cexsave = Rf_gpptr(dd)->cex;
     labelList = PROTECT(R_NilValue);
+
+
+    /* draw contour for levels[i] */
     GMode(1, dd);
-    for (i = 0; i < nc; i++) {/* draw contour for levels[i] */
+    for (i = 0; i < nc; i++) {
 	vmax = vmaxget();
 	Rf_gpptr(dd)->lty = INTEGER(lty)[i % nlty];
 	if (Rf_gpptr(dd)->lty == NA_INTEGER)
 	    Rf_gpptr(dd)->lty = ltysave;
-	Rf_gpptr(dd)->col = INTEGER(col)[i % ncol];
-	if (Rf_gpptr(dd)->col == NA_INTEGER)
+	if (isNAcol(rawcol, i, ncol))
 	    Rf_gpptr(dd)->col = colsave;
+	else
+	    Rf_gpptr(dd)->col = INTEGER(col)[i % ncol];
 	Rf_gpptr(dd)->lwd = REAL(lwd)[i % nlwd];
-	if (Rf_gpptr(dd)->lwd == NA_REAL)
+	if (!R_FINITE(Rf_gpptr(dd)->lwd))
 	    Rf_gpptr(dd)->lwd = lwdsave;
 	Rf_gpptr(dd)->cex = labcex;
 	contour(x, nx, y, ny, z, REAL(c)[i], labels, i,
@@ -1234,9 +1540,9 @@ SEXP do_contour(SEXP call, SEXP op, SEXP args, SEXP env)
     UNPROTECT(5);
     /* NOTE: only record operation if no "error"  */
     /* NOTE: on replay, call == R_NilValue */
-    if (GRecording(call))
+    if (GRecording(call, dd))
 	recordGraphicOperation(op, oargs, dd);
-    return R_NilValue;
+    return result;
 }
 
 
@@ -1270,13 +1576,6 @@ FindCutPoints(double low, double high,
 	    ++*npt;
 	}
 	if (z2 > low) {
-#ifdef OMIT
-	    /* Don't repeat corner vertices */
-	    x[*npt] = x2;
-	    y[*npt] = y2;
-	    z[*npt] = z2;
-	    ++*npt;
-#endif
 	}
 	else {
 	    c = (z2 -low) / (z2 - z1);
@@ -1394,15 +1693,15 @@ SEXP do_filledcontour(SEXP call, SEXP op, SEXP args, SEXP env)
     args = CDR(args);
 
     if (nx < 2 || ny < 2)
-	errorcall(call, "insufficient x or y values");
+	errorcall(call, _("insufficient 'x' or 'y' values"));
 
     if (nrows(sz) != nx || ncols(sz) != ny)
-	errorcall(call, "dimension mismatch");
+	errorcall(call, _("dimension mismatch"));
 
     if (nc < 1)
-	errorcall(call, "no contour values");
+	errorcall(call, _("no contour values"));
 
-    PROTECT(scol = FixupCol(CAR(args), NA_INTEGER));
+    PROTECT(scol = FixupCol(CAR(args), R_TRANWHITE));
     ncol = length(scol);
 
     /* Shorthand Pointers */
@@ -1451,7 +1750,7 @@ SEXP do_filledcontour(SEXP call, SEXP op, SEXP args, SEXP env)
 				    px, py, pz, &npt);
                 if (npt > 2)
 		    GPolygon(npt, px, py, USER, col[(k-1)%ncol],
-			     NA_INTEGER, dd);
+			     R_TRANWHITE, dd);
 	    }
 	}
     }
@@ -1460,14 +1759,14 @@ SEXP do_filledcontour(SEXP call, SEXP op, SEXP args, SEXP env)
     Rf_gpptr(dd)->xpd = xpdsave;
     R_Visible = 0;
     UNPROTECT(1);
-    if (GRecording(call))
+    if (GRecording(call, dd))
 	recordGraphicOperation(op, oargs, dd);
     return R_NilValue;
 
  badxy:
-    errorcall(call, "invalid x / y values or limits");
+    errorcall(call, _("invalid x / y values or limits"));
  badlev:
-    errorcall(call, "invalid contour levels: must be strictly increasing");
+    errorcall(call, _("invalid contour levels: must be strictly increasing"));
     return R_NilValue;  /* never used; to keep -Wall happy */
 }
 
@@ -1504,7 +1803,7 @@ SEXP do_image(SEXP call, SEXP op, SEXP args, SEXP env)
     internalTypeCheck(call, sz, INTSXP);
     args = CDR(args);
 
-    PROTECT(sc = FixupCol(CAR(args), NA_INTEGER));
+    PROTECT(sc = FixupCol(CAR(args), R_TRANWHITE));
     nc = LENGTH(sc);
 
     /* Shorthand Pointers */
@@ -1537,7 +1836,7 @@ SEXP do_image(SEXP call, SEXP op, SEXP args, SEXP env)
 	    tmp = z[i + j * (nx - 1)];
 	    if (tmp >= 0 && tmp < nc && tmp != NA_INTEGER)
 		GRect(x[i], y[j], x[i+1], y[j+1], USER, c[tmp],
-		      NA_INTEGER, dd);
+		      R_TRANWHITE, dd);
 	}
     }
     GMode(0, dd);
@@ -1545,30 +1844,18 @@ SEXP do_image(SEXP call, SEXP op, SEXP args, SEXP env)
     Rf_gpptr(dd)->xpd = xpdsave;
     R_Visible = 0;
     UNPROTECT(1);
-    if (GRecording(call))
+    if (GRecording(call, dd))
 	recordGraphicOperation(op, oargs, dd);
     return R_NilValue;
 
   badxy:
-    errorcall(call, "invalid x / y values or limits");
+    errorcall(call, _("invalid x / y values or limits"));
     return R_NilValue;/* never used; to keep -Wall happy */
 }
 
 	/*  P e r s p e c t i v e   S u r f a c e   P l o t s  */
 
 
-#ifdef NOT_used_currently/*-- out 'def'  (-Wall) --*/
-static void MakeVector (double x, double y, double z, Vector3d v)
-{
-    v[0] = x;
-    v[1] = y;
-    v[2] = z;
-    v[3] = 1;
-}
-#endif
-
-
-
 /* Set up the light source */
 static double Light[4];
 static double Shade;
@@ -1733,7 +2020,7 @@ static void CheckRange(double *x, int n, double min, double max)
 	    if (x[i] > xmax) xmax = x[i];
 	}
     if (xmin < min || xmax > max)
-	errorcall(gcall, "coordinates outsize specified range");
+	errorcall(gcall, _("coordinates outsize specified range"));
 }
 #endif
 
@@ -1958,7 +2245,7 @@ static void PerspAxis(double *x, double *y, double *z,
     i = nint;
     GPretty(&min, &max, &nint);
     /* GPretty() rarely gives values too much outside range ..
-       2D axis() clip these, we play cheaper */ 
+       2D axis() clip these, we play cheaper */
     while((min < range[0] - d_frac || range[1] + d_frac < max) && i < 20) {
 	nint = ++i;
 	min = range[0];
@@ -2151,7 +2438,7 @@ static void PerspAxes(double *x, double *y, double *z,
 	xAxis = 2;
 	yAxis = 3;
     } else
-	warning("Axis orientation not calculated");
+	warning(_("Axis orientation not calculated"));
     PerspAxis(x, y, z, xAxis, 0, nTicks, tickType, xlab, dd);
     PerspAxis(x, y, z, yAxis, 1, nTicks, tickType, ylab, dd);
     /* Figure out which Z axis to draw */
@@ -2164,7 +2451,7 @@ static void PerspAxes(double *x, double *y, double *z,
     } else if (lowest(v3[0]/v3[3], v1[0]/v1[3], v2[0]/v2[3], v0[0]/v0[3])) {
 	zAxis = 7;
     } else
-	warning("Axes orientation not calculated");
+	warning(_("Axis orientation not calculated"));
     PerspAxis(x, y, z, zAxis, 2, nTicks, tickType, zlab, dd);
 
     Rf_gpptr(dd)->xpd = xpdsave;
@@ -2182,43 +2469,43 @@ SEXP do_persp(SEXP call, SEXP op, SEXP args, SEXP env)
     DevDesc *dd;
 
     if (length(args) < 24)
-	errorcall(call, "too few parameters");
+	errorcall(call, _("too few parameters"));
     gcall = call;
     originalArgs = args;
 
     PROTECT(x = coerceVector(CAR(args), REALSXP));
-    if (length(x) < 2) errorcall(call, "invalid x argument");
+    if (length(x) < 2) errorcall(call, _("invalid 'x' argument"));
     args = CDR(args);
 
     PROTECT(y = coerceVector(CAR(args), REALSXP));
-    if (length(y) < 2) errorcall(call, "invalid y argument");
+    if (length(y) < 2) errorcall(call, _("invalid 'y' argument"));
     args = CDR(args);
 
     PROTECT(z = coerceVector(CAR(args), REALSXP));
     if (!isMatrix(z) || nrows(z) != length(x) || ncols(z) != length(y))
-	errorcall(call, "invalid z argument");
+	errorcall(call, _("invalid 'z' argument"));
     args = CDR(args);
 
     PROTECT(xlim = coerceVector(CAR(args), REALSXP));
-    if (length(xlim) != 2) errorcall(call, "invalid xlim argument");
+    if (length(xlim) != 2) errorcall(call, _("invalid 'xlim' argument"));
     args = CDR(args);
 
     PROTECT(ylim = coerceVector(CAR(args), REALSXP));
-    if (length(ylim) != 2) errorcall(call, "invalid ylim argument");
+    if (length(ylim) != 2) errorcall(call, _("invalid 'ylim' argument"));
     args = CDR(args);
 
     PROTECT(zlim = coerceVector(CAR(args), REALSXP));
-    if (length(zlim) != 2) errorcall(call, "invalid zlim argument");
+    if (length(zlim) != 2) errorcall(call, _("invalid 'zlim' argument"));
     args = CDR(args);
 
     /* Checks on x/y/z Limits */
 
     if (!LimitCheck(REAL(xlim), &xc, &xs))
-	errorcall(call, "invalid x limits");
+	errorcall(call, _("invalid 'x' limits"));
     if (!LimitCheck(REAL(ylim), &yc, &ys))
-	errorcall(call, "invalid y limits");
+	errorcall(call, _("invalid 'y' limits"));
     if (!LimitCheck(REAL(zlim), &zc, &zs))
-	errorcall(call, "invalid z limits");
+	errorcall(call, _("invalid 'z' limits"));
 
     theta = asReal(CAR(args));	args = CDR(args);
     phi	  = asReal(CAR(args));	args = CDR(args);
@@ -2239,11 +2526,11 @@ SEXP do_persp(SEXP call, SEXP op, SEXP args, SEXP env)
     ylab = CAR(args); args = CDR(args);
     zlab = CAR(args); args = CDR(args);
     if (!isString(xlab) || length(xlab) < 1)
-	error("`xlab' must be a character vector of length 1");
+	error(_("'xlab' must be a character vector of length 1"));
     if (!isString(ylab) || length(ylab) < 1)
-	error("`ylab' must be a character vector of length 1");
+	error(_("'ylab' must be a character vector of length 1"));
     if (!isString(zlab) || length(zlab) < 1)
-	error("`zlab' must be a character vector of length 1");
+	error(_("'zlab' must be a character vector of length 1"));
 
     if (R_FINITE(Shade) && Shade <= 0) Shade = 1;
     if (R_FINITE(ltheta) && R_FINITE(lphi) && R_FINITE(Shade))
@@ -2263,29 +2550,30 @@ SEXP do_persp(SEXP call, SEXP op, SEXP args, SEXP env)
 
     if (!R_FINITE(theta) || !R_FINITE(phi) || !R_FINITE(r) || !R_FINITE(d) ||
 	d < 0 || r < 0)
-	errorcall(call, "invalid viewing parameters");
+	errorcall(call, _("invalid viewing parameters"));
     if (!R_FINITE(expand) || expand < 0)
-	errorcall(call, "invalid expand value");
+	errorcall(call, _("invalid 'expand' value"));
     if (scale == NA_LOGICAL)
 	scale = 0;
     if ((nTicks == NA_INTEGER) || (nTicks < 0))
-	errorcall(call, "invalid nticks value");
+	errorcall(call, _("invalid 'nticks' value"));
     if ((tickType == NA_INTEGER) || (tickType < 1) || (tickType > 2))
-	errorcall(call, "invalid ticktype value");
+	errorcall(call, _("invalid 'ticktype' value"));
 
-    dd = GNewPlot(GRecording(call));
+    dd = CurrentDevice();
+
+    GNewPlot(GRecording(call, dd));
 
     PROTECT(col = FixupCol(col, Rf_gpptr(dd)->bg));
     ncol = LENGTH(col);
-    if (ncol < 1) errorcall(call, "invalid col specification");
+    if (ncol < 1) errorcall(call, _("invalid 'col' specification"));
     if(!R_OPAQUE(INTEGER(col)[0])) DoLighting = FALSE;
     PROTECT(border = FixupCol(border, Rf_gpptr(dd)->fg));
-    if (length(border) < 1) errorcall(call, "invalid border specification");
+    if (length(border) < 1) errorcall(call, _("invalid 'border' specification"));
 
     GSetState(1, dd);
     GSavePars(dd);
-    RecordGraphicsCall(call);
-    ProcessInlinePars(args, dd);
+    ProcessInlinePars(args, dd, call);
     if (length(border) > 1)
 	Rf_gpptr(dd)->fg = INTEGER(border)[0];
     Rf_gpptr(dd)->xlog = Rf_gpptr(dd)->ylog = FALSE;
@@ -2331,11 +2619,15 @@ SEXP do_persp(SEXP call, SEXP op, SEXP args, SEXP env)
 
     if (dobox) {
         PerspBox(0, REAL(xlim), REAL(ylim), REAL(zlim), dd);
-	if (doaxes)
+	if (doaxes) {
+	    SEXP xl = STRING_ELT(xlab, 0), yl = STRING_ELT(ylab, 0),
+		zl = STRING_ELT(zlab, 0);
 	    PerspAxes(REAL(xlim), REAL(ylim), REAL(zlim),
-		      CHAR(STRING_ELT(xlab, 0)), CHAR(STRING_ELT(ylab, 0)),
-		      CHAR(STRING_ELT(zlab, 0)),
+		      (xl == NA_STRING)? "" : CHAR(xl),
+		      (yl == NA_STRING)? "" : CHAR(yl),
+		      (zl == NA_STRING)? "" : CHAR(zl),
 		      nTicks, tickType, dd);
+	}
     }
 
     DrawFacets(REAL(z), REAL(x), REAL(y), nrows(z), ncols(z), INTEGER(indx),
@@ -2348,7 +2640,7 @@ SEXP do_persp(SEXP call, SEXP op, SEXP args, SEXP env)
 
     GRestorePars(dd);
     UNPROTECT(10);
-    if (GRecording(call))
+    if (GRecording(call, dd))
         recordGraphicOperation(op, originalArgs, dd);
 
     PROTECT(x = allocVector(REALSXP, 16));
