@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2002  R Development Core Team
+ *  Copyright (C) 1997--2005  R Development Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -18,6 +18,11 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/* <UTF8> char here is handled as a whole.
+   Collation is locale-specific if strcoll exists and works.
+ */
+
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -29,6 +34,7 @@ static SEXP integer_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
 static SEXP real_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
 static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call);
 static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
+static SEXP raw_relop(RELOP_TYPE code, SEXP s1, SEXP s2);
 
 SEXP do_relop(SEXP call, SEXP op, SEXP args, SEXP env)
 {
@@ -62,8 +68,8 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	if (nx > 0 && ny > 0)
 	    mismatch = ((nx > ny) ? nx % ny : ny % nx) != 0;
 	if (mismatch)
-	    warningcall(call, "longer object length\n"
-			"\tis not a multiple of shorter object length");
+	    warningcall(call, _("longer object length\n\
+ \tis not a multiple of shorter object length"));
 	UNPROTECT(2);
 	return ans;
     }
@@ -71,36 +77,34 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
     if ((iS = isSymbol(x)) || TYPEOF(x) == LANGSXP) {
 	SEXP tmp = allocVector(STRSXP, 1);
 	PROTECT(tmp);
-	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(x) : 
-		       STRING_ELT(deparse1(x, 0), 0));
+	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(x) :
+		       STRING_ELT(deparse1(x, 0, SIMPLEDEPARSE), 0));
 	REPROTECT(x = tmp, xpi);
 	UNPROTECT(1);
     }
     if ((iS = isSymbol(y)) || TYPEOF(y) == LANGSXP) {
 	SEXP tmp = allocVector(STRSXP, 1);
 	PROTECT(tmp);
-	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(y) : 
-		       STRING_ELT(deparse1(y, 0), 0));
+	SET_STRING_ELT(tmp, 0, (iS) ? PRINTNAME(y) :
+		       STRING_ELT(deparse1(y, 0, SIMPLEDEPARSE), 0));
 	REPROTECT(y = tmp, ypi);
 	UNPROTECT(1);
     }
 
-    /* FIXME (?): S does
-    if (!isVectorAtomic(x) || !isVectorAtomic(y)) { */
     if (!isVector(x) || !isVector(y)) {
 	if (isNull(x) || isNull(y)) {
 	    UNPROTECT(2);
 	    return allocVector(LGLSXP,0);
 	}
 	errorcall(call,
-		  "comparison (%d) is possible only for atomic types",
+		  _("comparison (%d) is possible only for atomic and list types"),
 		  PRIMVAL(op));
     }
 
     if (TYPEOF(x) == EXPRSXP || TYPEOF(y) == EXPRSXP)
-	errorcall(call, "comparison is not allowed for expressions");
+	errorcall(call, _("comparison is not allowed for expressions"));
 
-    /* ELSE :  x and y are both atomic, no not currently */
+    /* ELSE :  x and y are both atomic or list */
 
     if (LENGTH(x) <= 0 || LENGTH(y) <= 0) {
 	UNPROTECT(2);
@@ -114,11 +118,11 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
     yts = isTs(y);
     if (nx > 0 && ny > 0)
 	mismatch = ((nx > ny) ? nx % ny : ny % nx) != 0;
- 
+
     if (xarray || yarray) {
 	if (xarray && yarray) {
 	    if (!conformable(x, y))
-		errorcall(call, "non-conformable arrays");
+		errorcall(call, _("non-conformable arrays"));
 	    PROTECT(dims = getAttrib(x, R_DimSymbol));
 	}
 	else if (xarray) {
@@ -138,7 +142,7 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
     if (xts || yts) {
 	if (xts && yts) {
 	    if (!tsConform(x, y))
-		errorcall(call, "Non-conformable time-series");
+		errorcall(call, _("non-conformable time series"));
 	    PROTECT(tsp = getAttrib(x, R_TspSymbol));
 	    PROTECT(class = getAttrib(x, R_ClassSymbol));
 	}
@@ -156,8 +160,7 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	}
     }
     if (mismatch)
-	warningcall(call, "longer object length\n"
-		    "\tis not a multiple of shorter object length");
+	warningcall(call, _("longer object length\n\tis not a multiple of shorter object length"));
 
     if (isString(x) || isString(y)) {
 	REPROTECT(x = coerceVector(x, STRSXP), xpi);
@@ -174,10 +177,21 @@ SEXP do_relop_dflt(SEXP call, SEXP op, SEXP x, SEXP y)
 	REPROTECT(y = coerceVector(y, REALSXP), ypi);
 	x = real_relop(PRIMVAL(op), x, y);
     }
-    else if((isInteger(x) || isLogical(x)) &&
-	    (isInteger(y) || isLogical(y)))
+    else if (isInteger(x) || isInteger(y)) {
+	REPROTECT(x = coerceVector(x, INTSXP), xpi);
+	REPROTECT(y = coerceVector(y, INTSXP), ypi);
 	x = integer_relop(PRIMVAL(op), x, y);
-    else errorcall(call, "comparison of these types is not implemented");
+    }
+    else if (isLogical(x) || isLogical(y)) {
+	REPROTECT(x = coerceVector(x, LGLSXP), xpi);
+	REPROTECT(y = coerceVector(y, LGLSXP), ypi);
+	x = integer_relop(PRIMVAL(op), x, y);
+    }
+    else if (TYPEOF(x) == RAWSXP || TYPEOF(y) == RAWSXP) {
+	REPROTECT(x = coerceVector(x, RAWSXP), xpi);
+	REPROTECT(y = coerceVector(y, RAWSXP), ypi);
+	x = raw_relop(PRIMVAL(op), x, y);
+    } else errorcall(call, _("comparison of these types is not implemented"));
 
 
     PROTECT(x);
@@ -378,7 +392,7 @@ static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call)
     SEXP ans;
 
     if (code != EQOP && code != NEOP) {
-	errorcall(call, "illegal comparison with complex values");
+	errorcall(call, _("invalid comparison with complex values"));
     }
 
     n1 = LENGTH(s1);
@@ -418,10 +432,17 @@ static SEXP complex_relop(RELOP_TYPE code, SEXP s1, SEXP s2, SEXP call)
     UNPROTECT(2);
     return ans;
 }
-#ifdef HAVE_STRCOLL
-#define STRCMP strcoll
+
+#if defined(Win32) && defined(SUPPORT_UTF8)
+#define STRCOLL Rstrcoll
 #else
-#define STRCMP strcmp
+
+#ifdef HAVE_STRCOLL
+#define STRCOLL strcoll
+#else
+#define STRCOLL strcmp
+#endif
+
 #endif
 
 static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
@@ -469,7 +490,7 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 		(STRING_ELT(s2, i % n2) == NA_STRING))
 		LOGICAL(ans)[i] = NA_LOGICAL;
  	    else
-	    if (STRCMP(CHAR(STRING_ELT(s1, i % n1)),
+	    if (STRCOLL(CHAR(STRING_ELT(s1, i % n1)),
 		       CHAR(STRING_ELT(s2, i % n2))) < 0)
 		LOGICAL(ans)[i] = 1;
 	    else
@@ -482,7 +503,7 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 		(STRING_ELT(s2, i % n2) == NA_STRING))
 		LOGICAL(ans)[i] = NA_LOGICAL;
  	    else
-	    if (STRCMP(CHAR(STRING_ELT(s1, i % n1)),
+	    if (STRCOLL(CHAR(STRING_ELT(s1, i % n1)),
 		       CHAR(STRING_ELT(s2, i % n2))) > 0)
 		LOGICAL(ans)[i] = 1;
 	    else
@@ -495,7 +516,7 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 		(STRING_ELT(s2, i % n2) == NA_STRING))
 		LOGICAL(ans)[i] = NA_LOGICAL;
  	    else
-	    if (STRCMP(CHAR(STRING_ELT(s1, i % n1)),
+	    if (STRCOLL(CHAR(STRING_ELT(s1, i % n1)),
 		       CHAR(STRING_ELT(s2, i % n2))) <= 0)
 		LOGICAL(ans)[i] = 1;
 	    else
@@ -508,7 +529,7 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
 		(STRING_ELT(s2, i % n2) == NA_STRING))
 		LOGICAL(ans)[i] = NA_LOGICAL;
  	    else
-	    if (STRCMP(CHAR(STRING_ELT(s1, i % n1)),
+	    if (STRCOLL(CHAR(STRING_ELT(s1, i % n1)),
 		       CHAR(STRING_ELT(s2, i % n2))) >= 0)
 		LOGICAL(ans)[i] = 1;
 	    else
@@ -519,3 +540,64 @@ static SEXP string_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
     UNPROTECT(2);
     return ans;
 }
+
+static SEXP raw_relop(RELOP_TYPE code, SEXP s1, SEXP s2)
+{
+    int i, i1, i2, n, n1, n2;
+    Rbyte x1, x2;
+    SEXP ans;
+
+    n1 = LENGTH(s1);
+    n2 = LENGTH(s2);
+    n = (n1 > n2) ? n1 : n2;
+    PROTECT(s1);
+    PROTECT(s2);
+    ans = allocVector(LGLSXP, n);
+
+    switch (code) {
+    case EQOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 == x2);
+	}
+	break;
+    case NEOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 != x2);
+	}
+	break;
+    case LTOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 < x2);
+	}
+	break;
+    case GTOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 > x2);
+	}
+	break;
+    case LEOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 <= x2);
+	}
+	break;
+    case GEOP:
+	mod_iterate(n1, n2, i1, i2) {
+	    x1 = RAW(s1)[i1];
+	    x2 = RAW(s2)[i2];
+	    LOGICAL(ans)[i] = (x1 >= x2);
+	}
+	break;
+    }
+    UNPROTECT(2);
+    return ans;
+}
