@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2002  Robert Gentleman, Ross Ihaka and the
+ *  Copyright (C) 1997--2004  Robert Gentleman, Ross Ihaka and the
  *			      R Development Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -19,15 +19,12 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/* <UTF8> byte-level access is only to compare with chars <= 0x7F */
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
 
-/* we substitute if XDR is not found */
-#ifndef HAVE_XDR
-# define HAVE_XDR 1
-#endif
-
 #define NEED_CONNECTION_PSTREAMS
 #include <Defn.h>
 #include <Rmath.h>
@@ -158,7 +155,10 @@
 static void OutStringVec(R_outpstream_t stream, SEXP s, SEXP ref_table);
 static void WriteItem (SEXP s, SEXP ref_table, R_outpstream_t stream);
 static SEXP ReadItem(SEXP ref_table, R_inpstream_t stream);
-
+#ifdef BYTECODE
+static void WriteBC(SEXP s, SEXP ref_table, R_outpstream_t stream);
+static SEXP ReadBC(SEXP ref_table, R_inpstream_t stream);
+#endif
 
 /*
  * Constants
@@ -180,7 +180,7 @@ static int R_DefaultSerializeVersion = 2;
 #define R_assert(e) ((void) 0)
 #else
 /* The line below requires an ANSI C preprocessor (stringify operator) */
-#define R_assert(e) ((e) ? (void) 0 : error("assertion `%s' failed: file `%s', line %d\n", #e, __FILE__, __LINE__))
+#define R_assert(e) ((e) ? (void) 0 : error("assertion '%s' failed: file '%s', line %d\n", #e, __FILE__, __LINE__))
 #endif /* NDEBUG */
 
 /* Rsnprintf: like snprintf, but guaranteed to null-terminate. */
@@ -219,7 +219,7 @@ static void OutInteger(R_outpstream_t stream, int i)
 	stream->OutBytes(stream, buf, R_XDR_INTEGER_SIZE);
 	break;
     default:
-	error("unknown or inappropriate output format");
+	error(_("unknown or inappropriate output format"));
     }
 }
 
@@ -249,7 +249,7 @@ static void OutReal(R_outpstream_t stream, double d)
 	stream->OutBytes(stream, buf, R_XDR_DOUBLE_SIZE);
 	break;
     default:
-	error("unknown or inappropriate output format");
+	error(_("unknown or inappropriate output format"));
     }
 }
 
@@ -259,6 +259,23 @@ static void OutComplex(R_outpstream_t stream, Rcomplex c)
     OutReal(stream, c.i);
 }
 
+static void OutByte(R_outpstream_t stream, int i)
+{
+    char buf[128];
+    switch (stream->type) {
+    case R_pstream_ascii_format:
+	Rsnprintf(buf, sizeof(buf), "%02x\n", i);
+	stream->OutBytes(stream, buf, strlen(buf));
+	break;
+    case R_pstream_binary_format:
+    case R_pstream_xdr_format:
+	stream->OutBytes(stream, &i, 1);
+	break;
+    default:
+	error(_("unknown or inappropriate output format"));
+    }
+}
+
 static void OutString(R_outpstream_t stream, char *s, int length)
 {
     if (stream->type == R_pstream_ascii_format) {
@@ -308,14 +325,14 @@ static void InWord(R_inpstream_t stream, char * buf, int size)
     do {
 	c = stream->InChar(stream);
 	if (c == EOF)
-	    error("read error");
+	    error(_("read error"));
     } while (isspace(c));
     while (! isspace(c) && i < size) {
 	buf[i++] = c;
 	c = stream->InChar(stream);
     }
     if (i == size)
-	error("read error");
+	error(_("read error"));
     buf[i] = 0;
 }
     
@@ -383,6 +400,24 @@ static Rcomplex InComplex(R_inpstream_t stream)
     return c;
 }
 
+static int InByte(R_inpstream_t stream)
+{
+    char word[128];
+    Rbyte rb;
+
+    switch (stream->type) {
+    case R_pstream_ascii_format:
+	InWord(stream, word, sizeof(word));
+	return (Rbyte) word[0];
+    case R_pstream_binary_format:
+    case R_pstream_xdr_format:
+	stream->InBytes(stream, &rb, 1);
+	return rb;
+    default:
+	return 0;
+    }
+}
+
 /* These utilities for reading characters with an unget option are
    defined so the code in InString can match the code in
    saveload.c:InStringAscii--that way it is easier to match changes in
@@ -473,21 +508,16 @@ static void InString(R_inpstream_t stream, char *buf, int length)
 static void OutFormat(R_outpstream_t stream)
 {
     if (stream->type == R_pstream_binary_format) {
-#ifdef HAVE_XDR
-	warning("binary format is deprecated; using xdr instead");
+	warning(_("binary format is deprecated; using xdr instead"));
 	stream->type = R_pstream_xdr_format;
-#else
-	warning("portable binary format is not available; using ascii");
-	stream->type = R_pstream_ascii_format;
-#endif
     }
     switch (stream->type) {
     case R_pstream_ascii_format:  stream->OutBytes(stream, "A\n", 2); break;
     case R_pstream_binary_format: stream->OutBytes(stream, "B\n", 2); break;
     case R_pstream_xdr_format:    stream->OutBytes(stream, "X\n", 2); break;
     case R_pstream_any_format:
-	error("must specify ascii, binary, or xdr format");
-    default: error("unknown output format");
+	error(_("must specify ascii, binary, or xdr format"));
+    default: error(_("unknown output format"));
     }
 }
 
@@ -500,21 +530,37 @@ static void InFormat(R_inpstream_t stream)
     case 'A': type = R_pstream_ascii_format; break;
     case 'B': type = R_pstream_binary_format; break;
     case 'X': type = R_pstream_xdr_format; break;
+    case '\n':
+	/* GROSS HACK: ASCII unserialize may leave a trailing newline
+	   in the stream.  If the stream contains a second
+	   serialization, then a second unserialize will fail if such
+	   a newline is present.  The right fix is to make sure
+	   unserialize consumes exactly what serialize produces.  But
+	   this seems hard because of the current use of whitespace
+	   skipping in unserialize.  So a temporary hack to cure the
+	   symptom is to deal with a possible leading newline.  I
+	   don't think more than one is possible, but I'm not sure.
+	   LT */
+	if (buf[1] == 'A') {
+	    type = R_pstream_ascii_format;
+	    stream->InBytes(stream, buf, 1);
+	    break;
+	}
     default:
 	type = R_pstream_any_format;  /* keep compiler happy */
-	error("unknown input format");
+	error(_("unknown input format"));
     }
     if (stream->type == R_pstream_any_format)
 	stream->type = type;
     else if (type != stream->type)
-	error("input format does not match specified format");
+	error(_("input format does not match specified format"));
 }
 
 
 /*
  * Hash Table Functions
  *
- * Hashing functions for hashing reference objects diring writing.
+ * Hashing functions for hashing reference objects during writing.
  * Objects are entered, and the order in which they are encountered is
  * recorded.  GashGet returns this number, a positive integer, if the
  * object was seen before, and zero if not.  A fixed hash table size
@@ -588,6 +634,10 @@ static int HashGet(SEXP item, SEXP ht)
 /* the following are speculative--we may or may not need them soon */
 #define CLASSREFSXP       246
 #define GENERICREFSXP     245
+#ifdef BYTECODE
+#define BCREPDEF          244
+#define BCREPREF          243
+#endif
 
 /*
  * Type/Flag Packing and Unpacking
@@ -674,9 +724,7 @@ static SEXP GetPersistentName(R_outpstream_t stream, SEXP s)
 	case EXTPTRSXP: break;
 	case ENVSXP:
 	    if (s == R_GlobalEnv ||
-#ifdef EXPERIMENTAL_NAMESPACES
 		R_IsNamespaceEnv(s) ||
-#endif
 		R_IsPackageEnv(s))
 		return R_NilValue;
 	    else
@@ -692,7 +740,7 @@ static SEXP GetPersistentName(R_outpstream_t stream, SEXP s)
 static SEXP PersistentRestore(R_inpstream_t stream, SEXP s)
 {
     if (stream->InPersistHookFunc == NULL)
-	error("no restore method available");
+	error(_("no restore method available"));
     return stream->InPersistHookFunc(s, stream->InPersistHookData);
 }
 
@@ -707,9 +755,7 @@ static int SaveSpecialHook(SEXP item)
     if (item == R_GlobalEnv)     return GLOBALENV_SXP;
     if (item == R_UnboundValue)  return UNBOUNDVALUE_SXP;
     if (item == R_MissingArg)    return MISSINGARG_SXP;
-#ifdef EXPERIMENTAL_NAMESPACES
     if (item == R_BaseNamespace) return BASENAMESPACE_SXP;
-#endif
     return 0;
 }
 
@@ -721,8 +767,10 @@ static void OutStringVec(R_outpstream_t stream, SEXP s, SEXP ref_table)
     R_assert(TYPEOF(s) == STRSXP);
 
     names = getAttrib(s, R_NamesSymbol);
+#ifdef WARN_ABOUT_NAMES_IN_PERSISTENT_STRINGS
     if (names != R_NilValue)
-	warning("names in persistent strings are currently ignored");
+	warning(_("names in persistent strings are currently ignored"));
+#endif
 
     len = LENGTH(s);
     OutInteger(stream, 0); /* place holder to allow names if we want to */
@@ -743,6 +791,7 @@ static void OutStringVec(R_outpstream_t stream, SEXP s, SEXP ref_table)
 #define INTEGER_ELT(x,__i__)	INTEGER(x)[__i__]
 #define REAL_ELT(x,__i__)	REAL(x)[__i__]
 #define COMPLEX_ELT(x,__i__)	COMPLEX(x)[__i__]
+#define RAW_ELT(x,__i__)	RAW(x)[__i__]
 
 static void WriteItem (SEXP s, SEXP ref_table, R_outpstream_t stream)
 {
@@ -772,25 +821,21 @@ static void WriteItem (SEXP s, SEXP ref_table, R_outpstream_t stream)
 	HashAdd(s, ref_table);
 	if (R_IsPackageEnv(s)) {
 	    SEXP name = R_PackageEnvName(s);
-	    warning("%s may not be available when loading",
+	    warning(_("'%s' may not be available when loading"),
 		    CHAR(STRING_ELT(name, 0)));
 	    OutInteger(stream, PACKAGESXP);
 	    OutStringVec(stream, name, ref_table);
 	}
-#ifdef EXPERIMENTAL_NAMESPACES
 	else if (R_IsNamespaceEnv(s)) {
-	    warning("namespaces may not be available when loading");
-	    OutInteger(stream, NAMESPACESXP);
-	    OutStringVec(stream, R_NamespaceEnvName(s), ref_table);
-	}
+#ifdef WARN_ABOUT_NAME_SPACES_MAYBE_NOT_AVAILABLE
+	    warning(_("namespaces may not be available when loading"));
 #endif
+	    OutInteger(stream, NAMESPACESXP);
+	    OutStringVec(stream, R_NamespaceEnvSpec(s), ref_table);
+	}
 	else {
 	    OutInteger(stream, ENVSXP);
-#ifdef ENVIRONMENT_LOCKING
 	    OutInteger(stream, R_EnvironmentIsLocked(s) ? 1 : 0);
-#else
-	    OutInteger(stream, 0);
-#endif
 	    WriteItem(ENCLOS(s), ref_table, stream);
 	    WriteItem(FRAME(s), ref_table, stream);
 	    WriteItem(TAG(s), ref_table, stream);
@@ -798,8 +843,17 @@ static void WriteItem (SEXP s, SEXP ref_table, R_outpstream_t stream)
 	}
     }
     else {
-	int flags = PackFlags(TYPEOF(s), LEVELS(s), OBJECT(s),
-			      ATTRIB(s) != R_NilValue, TAG(s) != R_NilValue);
+	int flags, hastag;
+	switch(TYPEOF(s)) {
+	case LISTSXP:
+	case LANGSXP:
+	case CLOSXP:
+	case PROMSXP:
+	case DOTSXP: hastag = TAG(s) != R_NilValue; break;
+	default: hastag = FALSE;
+	}
+	flags = PackFlags(TYPEOF(s), LEVELS(s), OBJECT(s),
+                             ATTRIB(s) != R_NilValue, hastag);
 	OutInteger(stream, flags);
 	switch (TYPEOF(s)) {
 	case LISTSXP:
@@ -862,14 +916,175 @@ static void WriteItem (SEXP s, SEXP ref_table, R_outpstream_t stream)
 	    for (i = 0; i < LENGTH(s); i++)
 		WriteItem(VECTOR_ELT(s, i), ref_table, stream);
 	    break;
+	case BCODESXP:
+#ifdef BYTECODE
+	    WriteBC(s, ref_table, stream);
+	    break;
+#else
+	    error(_("this version of R cannot write byte code objects"));
+#endif
+	case RAWSXP:
+	    OutInteger(stream, LENGTH(s));
+	    OutVec(stream, s, RAW_ELT, OutByte);
+	    break;
 	default:
-	    error("WriteItem: unknown type %i", TYPEOF(s));
+	    error(_("WriteItem: unknown type %i"), TYPEOF(s));
 	}
 	if (ATTRIB(s) != R_NilValue)
 	    WriteItem(ATTRIB(s), ref_table, stream);
     }
 }
 
+#ifdef BYTECODE
+static SEXP MakeCircleHashTable()
+{
+    return CONS(R_NilValue, allocVector(VECSXP, HASHSIZE));
+}
+
+static Rboolean AddCircleHash(SEXP item, SEXP ct)
+{
+    SEXP table, bucket, list;
+    int pos;
+
+    table = CDR(ct);
+    pos = PTRHASH(item) % LENGTH(table);
+    bucket = VECTOR_ELT(table, pos);
+    for (list = bucket; list != R_NilValue; list = CDR(list))
+	if (TAG(list) == item) {
+	    if (CAR(list) == R_NilValue) {
+		/* this is the second time; enter in list and mark */
+		SETCAR(list, R_UnboundValue); /* anything different will do */
+		SETCAR(ct, CONS(item, CAR(ct)));
+	    }
+	    return TRUE;
+	}
+
+    /* If we get here then this is a new item; enter in the table */
+    bucket = CONS(R_NilValue, bucket);
+    SET_TAG(bucket, item);
+    SET_VECTOR_ELT(table, pos, bucket);
+    return FALSE;
+}
+
+static void ScanForCircles1(SEXP s, SEXP ct)
+{
+    switch (TYPEOF(s)) {
+    case LANGSXP:
+    case LISTSXP:
+        if (! AddCircleHash(s, ct)) {
+	    ScanForCircles1(CAR(s), ct);
+	    ScanForCircles1(CDR(s), ct);
+	}
+	break;
+    case BCODESXP:
+	{
+	    int i, n;
+	    SEXP consts = BCODE_CONSTS(s);
+	    n = LENGTH(consts);
+	    for (i = 0; i < n; i++)
+		ScanForCircles1(VECTOR_ELT(consts, i), ct);
+	}
+	break;
+    default: break;
+    }
+}
+
+static SEXP ScanForCircles(SEXP s)
+{
+    SEXP ct;
+    PROTECT(ct = MakeCircleHashTable());
+    ScanForCircles1(s, ct);
+    UNPROTECT(1);
+    return CAR(ct);
+}
+
+static SEXP findrep(SEXP x, SEXP reps)
+{
+    for (; reps != R_NilValue; reps = CDR(reps))
+	if (x == CAR(reps))
+	    return reps;
+    return R_NilValue;
+}
+
+static void WriteBCLang(SEXP s, SEXP ref_table, SEXP reps,
+			R_outpstream_t stream)
+{
+    int type = TYPEOF(s);
+    if (type == LANGSXP || type == LISTSXP) {
+	SEXP r = findrep(s, reps);
+	int output = TRUE;
+	if (r != R_NilValue) {
+	    /* we have a cell referenced more than once */
+	    if (TAG(r) == R_NilValue) {
+		/* this is the first reference, so update and register
+                   the counter */
+		int i = INTEGER(CAR(reps))[0]++;
+		SET_TAG(r, allocVector(INTSXP, 1));
+		INTEGER(TAG(r))[0] = i;
+		OutInteger(stream, BCREPDEF);
+		OutInteger(stream, i);
+	    }
+	    else {
+		/* we've seen it before, so just put out the index */
+		OutInteger(stream, BCREPREF);
+		OutInteger(stream, INTEGER(TAG(r))[0]);
+		output = FALSE;
+	    }
+	}
+	if (output) {
+	    OutInteger(stream, type);
+	    WriteItem(TAG(s), ref_table, stream);
+	    WriteBCLang(CAR(s), ref_table, reps, stream);
+	    WriteBCLang(CDR(s), ref_table, reps, stream);
+	}
+    }
+    else {
+	OutInteger(stream, 0); /* pad */
+	WriteItem(s, ref_table, stream);
+    }
+}
+
+static void WriteBC1(SEXP s, SEXP ref_table, SEXP reps, R_outpstream_t stream)
+{
+    int i, n;
+    SEXP code, consts;
+    PROTECT(code = R_bcDecode(BCODE_CODE(s)));
+    WriteItem(code, ref_table, stream);
+    consts = BCODE_CONSTS(s);
+    n = LENGTH(consts);
+    OutInteger(stream, n);
+    for (i = 0; i < n; i++) {
+	SEXP c = VECTOR_ELT(consts, i);
+	int type = TYPEOF(c);
+	switch (type) {
+	case BCODESXP:
+	    OutInteger(stream, type);
+	    WriteBC1(c, ref_table, reps, stream);
+	    break;
+	case LANGSXP:
+	case LISTSXP:
+	    WriteBCLang(c, ref_table, reps, stream);
+	    break;
+	default:
+	    OutInteger(stream, type);
+	    WriteItem(c, ref_table, stream);
+	}
+    }
+    UNPROTECT(1);
+}
+
+static void WriteBC(SEXP s, SEXP ref_table, R_outpstream_t stream)
+{
+    SEXP reps = ScanForCircles(s);
+    PROTECT(reps = CONS(R_NilValue, reps));
+    OutInteger(stream, length(reps));
+    SETCAR(reps, allocVector(INTSXP, 1));
+    INTEGER(CAR(reps))[0] = 0; 
+    WriteBC1(s, ref_table, reps, stream);
+    UNPROTECT(1);
+}
+#endif
+
 void R_Serialize(SEXP s, R_outpstream_t stream)
 {
     SEXP ref_table;
@@ -883,7 +1098,7 @@ void R_Serialize(SEXP s, R_outpstream_t stream)
 	OutInteger(stream, R_VERSION);
 	OutInteger(stream, R_Version(1,4,0));
 	break;
-    default: error("version %d not supported", version);
+    default: error(_("version %d not supported"), version);
     }
 
     PROTECT(ref_table = MakeHashTable());
@@ -911,7 +1126,7 @@ static SEXP GetReadRef(SEXP table, int index)
     SEXP data = CAR(table);
 
     if (i < 0 || i >= LENGTH(data))
-	error("reference index out of range");
+	error(_("reference index out of range"));
     return VECTOR_ELT(data, i);
 }
 
@@ -941,7 +1156,7 @@ static SEXP InStringVec(R_inpstream_t stream, SEXP ref_table)
     SEXP s;
     int i, len;
     if (InInteger(stream) != 0)
-	error("names in persistent strings are not supported yet");
+	error(_("names in persistent strings are not supported yet"));
     len = InInteger(stream);
     PROTECT(s = allocVector(STRSXP, len));
     for (i = 0; i < len; i++)
@@ -963,6 +1178,7 @@ static SEXP InStringVec(R_inpstream_t stream, SEXP ref_table)
 #define SET_INTEGER_ELT(x,__i__,v)	(INTEGER_ELT(x,__i__)=(v))
 #define SET_REAL_ELT(x,__i__,v)		(REAL_ELT(x,__i__)=(v))
 #define SET_COMPLEX_ELT(x,__i__,v)	(COMPLEX_ELT(x,__i__)=(v))
+#define SET_RAW_ELT(x,__i__,v)		(RAW_ELT(x,__i__)=(v))
 
 static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
 {
@@ -981,13 +1197,7 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
     case UNBOUNDVALUE_SXP:  return R_UnboundValue;
     case MISSINGARG_SXP:    return R_MissingArg;
     case BASENAMESPACE_SXP:
-#ifdef EXPERIMENTAL_NAMESPACES
 	return R_BaseNamespace;
-#else
-	warning("base namespace not available in this verison;"
-		" using .GlobalEnv");
-	return R_GlobalEnv;
-#endif
     case REFSXP:
 	return GetReadRef(ref_table, InRefIndex(stream, flags));
     case PERSISTSXP:
@@ -1009,14 +1219,10 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
 	AddReadRef(ref_table, s);
 	return s;
     case NAMESPACESXP:
-#ifdef EXPERIMENTAL_NAMESPACES
 	PROTECT(s = InStringVec(stream, ref_table));
 	s = R_FindNamespace(s);
 	AddReadRef(ref_table, s);
 	UNPROTECT(1);
-#else
-	error("namespaces not available in this version");
-#endif
 	return s;
     case ENVSXP:
 	{
@@ -1032,10 +1238,13 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
 	    SET_FRAME(s, ReadItem(ref_table, stream));
 	    SET_TAG(s, ReadItem(ref_table, stream));
 	    SET_ATTRIB(s, ReadItem(ref_table, stream));
+	    if (ATTRIB(s) != R_NilValue &&
+		getAttrib(s, R_ClassSymbol) != R_NilValue)
+		/* We don't write out the object bit for environments,
+		   so reconstruct it here if needed. */
+		SET_OBJECT(s, 1);
 	    R_RestoreHashCount(s);
-#ifdef ENVIRONMENT_LOCKING
 	    if (locked) R_LockEnvironment(s, FALSE);
-#endif
 	    UNPROTECT(1);
 	    return s;
 	}
@@ -1062,7 +1271,7 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
 	UNPROTECT(1); /* s */
 	return s;
     default:
-	/* These break out of the swith to have their ATTR,
+	/* These break out of the switch to have their ATTR,
 	   LEVELS, and OBJECT fields filled in.  Each leaves the
 	   newly allocated value PROTECTed */
 	switch (type) {
@@ -1121,14 +1330,24 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
 		SET_VECTOR_ELT(s, count, ReadItem(ref_table, stream));
 	    break;
 	case BCODESXP:
-	    error("this version of R cannot read byte code objects");
+#ifdef BYTECODE
+	    PROTECT(s = ReadBC(ref_table, stream));
+	    break;
+#else
+	    error(_("this version of R cannot read byte code objects"));
+#endif
 	case CLASSREFSXP:
-	    error("this version of R cannot read class references");
+	    error(_("this version of R cannot read class references"));
 	case GENERICREFSXP:
-	    error("this version of R cannot read generic function references");
+	    error(_("this version of R cannot read generic function references"));
+	case RAWSXP:
+	    length = InInteger(stream);
+	    PROTECT(s = allocVector(type, length));
+	    InVec(stream, s, SET_RAW_ELT, InByte, length);
+	    break;
 	default:
 	    s = R_NilValue; /* keep compiler happy */
-	    error("ReadItem: unknown type %i", type);
+	    error(_("ReadItem: unknown type %i"), type);
 	}
 	SETLEVELS(s, levs);
 	SET_OBJECT(s, objf);
@@ -1138,6 +1357,90 @@ static SEXP ReadItem (SEXP ref_table, R_inpstream_t stream)
     }
 }
 
+#ifdef BYTECODE
+static SEXP ReadBC1(SEXP ref_table, SEXP reps, R_inpstream_t stream);
+
+static SEXP ReadBCLang(int type, SEXP ref_table, SEXP reps,
+		       R_inpstream_t stream)
+{
+    switch (type) {
+    case BCREPREF:
+	return VECTOR_ELT(reps, InInteger(stream));
+    case BCREPDEF:
+    case LANGSXP:
+    case LISTSXP:
+	{
+	    SEXP ans;
+	    int pos = -1;
+	    if (type == BCREPDEF) {
+		pos = InInteger(stream);
+		type = InInteger(stream);
+	    }
+	    PROTECT(ans = allocSExp(type));
+	    if (pos >= 0)
+		SET_VECTOR_ELT(reps, pos, ans);
+	    SET_TAG(ans, ReadItem(ref_table, stream));
+	    SETCAR(ans, ReadBCLang(InInteger(stream), ref_table, reps,
+				   stream));
+	    SETCDR(ans, ReadBCLang(InInteger(stream), ref_table, reps,
+				   stream));
+	    UNPROTECT(1);
+	    return ans;
+	}
+    default: return ReadItem(ref_table, stream);
+    }
+}
+
+static SEXP ReadBCConsts(SEXP ref_table, SEXP reps, R_inpstream_t stream)
+{
+    SEXP ans, c;
+    int i, n;
+    n = InInteger(stream);
+    PROTECT(ans = allocVector(VECSXP, n));
+    for (i = 0; i < n; i++) {
+	int type = InInteger(stream);
+	switch (type) {
+	case BCODESXP:
+	    c = ReadBC1(ref_table, reps, stream);
+	    SET_VECTOR_ELT(ans, i, c);
+	    break;
+	case LANGSXP:
+	case LISTSXP:
+	case BCREPDEF:
+	case BCREPREF:
+	    c = ReadBCLang(type, ref_table, reps, stream);
+	    SET_VECTOR_ELT(ans, i, c);
+	    break;
+	default:
+	    SET_VECTOR_ELT(ans, i, ReadItem(ref_table, stream));
+	}
+    }
+    UNPROTECT(1);
+    return ans;
+}
+    
+static SEXP ReadBC1(SEXP ref_table, SEXP reps, R_inpstream_t stream)
+{
+    SEXP s;
+    PROTECT(s = allocSExp(BCODESXP));
+    SETCAR(s, ReadItem(ref_table, stream)); /* code */
+    SETCAR(s, R_bcEncode(CAR(s)));
+    SETCDR(s, ReadBCConsts(ref_table, reps, stream)); /* consts */
+    SET_TAG(s, R_NilValue); /* expr */
+    UNPROTECT(1);
+    return s;
+}
+
+static SEXP ReadBC(SEXP ref_table, R_inpstream_t stream)
+{
+    SEXP reps, ans;
+    PROTECT(reps = allocVector(VECSXP, InInteger(stream)));
+    ans = ReadBC1(ref_table, reps, stream);
+    UNPROTECT(1);
+    return ans;
+}
+#endif
+
 static void DecodeVersion(int packed, int *v, int *p, int *s)
 {
     *v = packed / 65536; packed = packed % 65536;
@@ -1164,13 +1467,11 @@ SEXP R_Unserialize(R_inpstream_t stream)
 	    int vw, pw, sw;
 	    DecodeVersion(writer_version, &vw, &pw, &sw);
 	    if (release_version < 0)
-		error("can't read unreleased workspace version %d written by"
-		      " experimental R %d.%d.%d", version, vw, pw, sw);
+		error(_("cannot read unreleased workspace version %d written by experimental R %d.%d.%d"), version, vw, pw, sw);
 	    else {
 		int vm, pm, sm;
 		DecodeVersion(release_version, &vm, &pm, &sm);
-		error("can't read workspace version %d written by R %d.%d.%d;"
-		      " need R %d.%d.%d or newer",
+		error(_("cannot read workspace version %d written by R %d.%d.%d; need R %d.%d.%d or newer"),
 		      version, vw, pw, sw, vm, pm, sm);
 	    }
 	}
@@ -1274,17 +1575,17 @@ void R_InitFileInPStream(R_inpstream_t stream, FILE *fp,
 static void CheckInConn(Rconnection con)
 {
     if (! con->isopen)
-	error("connection is not open");
+	error(_("connection is not open"));
     if (! con->canread || con->read == NULL)
-	error("cannot read from this connection");
+	error(_("cannot read from this connection"));
 }
 
 static void CheckOutConn(Rconnection con)
 {
     if (! con->isopen)
-	error("connection is not open");
+	error(_("connection is not open"));
     if (! con->canwrite || con->write == NULL)
-	error("cannot write to this connection");
+	error(_("cannot write to this connection"));
 }
 
 static void InBytesConn(R_inpstream_t stream, void *buf, int length)
@@ -1299,7 +1600,7 @@ static void InBytesConn(R_inpstream_t stream, void *buf, int length)
     }
     else {
 	if (length != con->read(buf, 1, length, con))
-	    error("error reading from connection");
+	    error(_("error reading from connection"));
     }
 }
 
@@ -1312,7 +1613,7 @@ static int InCharConn(R_inpstream_t stream)
 	return Rconn_fgetc(con);
     else {
 	if (1 != con->read(buf, 1, 1, con))
-	    error("error reading from connection");
+	    error(_("error reading from connection"));
 	return buf[0];
     }
 }
@@ -1329,7 +1630,7 @@ static void OutBytesConn(R_outpstream_t stream, void *buf, int length)
     }
     else {
 	if (length != con->write(buf, 1, length, con))
-	    error("error writing to connection");
+	    error(_("error writing to connection"));
     }
 }
 
@@ -1343,7 +1644,7 @@ static void OutCharConn(R_outpstream_t stream, int c)
 	char buf[1];
 	buf[0] = (char) c;
 	if (1 != con->write(buf, 1, 1, con))
-	    error("error writing to connection");
+	    error(_("error writing to connection"));
     }
 }
 
@@ -1353,7 +1654,7 @@ void R_InitConnOutPStream(R_outpstream_t stream, Rconnection con,
 {
     CheckOutConn(con);
     if (con->text && type != R_pstream_ascii_format)
-	error("only ascii format can be written to text mode connections");
+	error(_("only ascii format can be written to text mode connections"));
     R_InitOutPStream(stream, (R_pstream_data_t) con, type, version,
 		     OutCharConn, OutBytesConn, phook, pdata);
 }
@@ -1367,8 +1668,535 @@ void R_InitConnInPStream(R_inpstream_t stream,  Rconnection con,
 	if (type == R_pstream_any_format)
 	    type = R_pstream_ascii_format;
 	else if (type != R_pstream_ascii_format)
-	    error("only ascii format can be read from text mode connections");
+	    error(_("only ascii format can be read from text mode connections"));
     }
     R_InitInPStream(stream, (R_pstream_data_t) con, type,
 		    InCharConn, InBytesConn, phook, pdata);
 }
+
+/* ought to quote the argument, but it should only be an ENVSXP or STRSXP */
+static SEXP CallHook(SEXP x, SEXP fun)
+{
+    SEXP val, call;
+    PROTECT(call = LCONS(fun, LCONS(x, R_NilValue)));
+    val = eval(call, R_GlobalEnv);
+    UNPROTECT(1);
+    return val;
+}
+
+SEXP do_serializeToConn(SEXP call, SEXP op, SEXP args, SEXP env)
+{
+    /* serializeToConn(object, conn, ascii, version, hook) */
+
+    SEXP object, fun;
+    Rboolean ascii;
+    int version;
+    Rconnection con;
+    struct R_outpstream_st out;
+    R_pstream_format_t type;
+    SEXP (*hook)(SEXP, SEXP);
+
+    checkArity(op, args);
+
+    object = CAR(args);
+    con = getConnection(asInteger(CADR(args)));
+
+    if (TYPEOF(CADDR(args)) != LGLSXP)
+	errorcall(call, _("'ascii' must be logical"));
+    ascii = INTEGER(CADDR(args))[0];
+    if (ascii) type = R_pstream_ascii_format;
+    else type = R_pstream_xdr_format;
+
+    if (CADDDR(args) == R_NilValue)
+	version = R_DefaultSerializeVersion;
+    else
+	version = asInteger(CADDDR(args));
+    if (version == NA_INTEGER || version <= 0)
+	error(_("bad version value"));
+    if (version < 2)
+	error(_("cannot save to connections in version %d format"), version);
+
+    fun = CAR(nthcdr(args,4));
+    hook = fun != R_NilValue ? CallHook : NULL;
+
+    R_InitConnOutPStream(&out, con, type, version, hook, fun);
+    R_Serialize(object, &out);
+    return R_NilValue;
+}
+
+SEXP do_unserializeFromConn(SEXP call, SEXP op, SEXP args, SEXP env)
+{
+    /* unserializeFromConn(conn, hook) */
+
+    struct R_inpstream_st in;
+    Rconnection con;
+    SEXP fun;
+    SEXP (*hook)(SEXP, SEXP);
+
+    checkArity(op, args);
+
+    con = getConnection(asInteger(CAR(args)));
+
+    fun = CADR(args);
+    hook = fun != R_NilValue ? CallHook : NULL;
+
+    R_InitConnInPStream(&in, con, R_pstream_any_format, hook, fun);
+    return R_Unserialize(&in);
+}
+
+
+/*
+ * Persistent Buffered Binary Connection Streams
+ */
+
+/**** should eventually come from a public header file */
+size_t R_WriteConnection(Rconnection con, void *buf, size_t n);
+
+#define BCONBUFSIZ 4096
+
+typedef struct bconbuf_st {
+    Rconnection con;
+    int count;
+    unsigned char buf[BCONBUFSIZ];
+} *bconbuf_t;
+
+static void flush_bcon_buffer(bconbuf_t bb)
+{
+    if (R_WriteConnection(bb->con, bb->buf, bb->count) != bb->count)
+	error(_("error writing to connection"));
+    bb->count = 0;
+}
+
+static void OutCharBB(R_outpstream_t stream, int c)
+{
+    bconbuf_t bb = stream->data;
+    if (bb->count >= BCONBUFSIZ)
+	flush_bcon_buffer(bb);
+    bb->buf[bb->count++] = c;
+}
+
+static void OutBytesBB(R_outpstream_t stream, void *buf, int length)
+{
+    bconbuf_t bb = stream->data;
+    if (bb->count + length > BCONBUFSIZ)
+	flush_bcon_buffer(bb);
+    if (length <= BCONBUFSIZ) {
+	memcpy(bb->buf + bb->count, buf, length);
+	bb->count += length;
+    }
+    else if (R_WriteConnection(bb->con, buf, length) != length)
+	error(_("error writing to connection"));
+}
+
+static void InitBConOutPStream(R_outpstream_t stream, bconbuf_t bb,
+			       Rconnection con,
+			       R_pstream_format_t type, int version,
+			       SEXP (*phook)(SEXP, SEXP), SEXP pdata)
+{
+    bb->count = 0;
+    bb->con = con;
+    R_InitOutPStream(stream, (R_pstream_data_t) bb, type, version,
+		     OutCharBB, OutBytesBB, phook, pdata);
+}
+
+SEXP R_serializeb(SEXP object, SEXP icon, SEXP fun)
+{
+    struct R_outpstream_st out;
+    SEXP (*hook)(SEXP, SEXP);
+    struct bconbuf_st bbs;
+    Rconnection con = getConnection(asInteger(icon));
+
+    hook = fun != R_NilValue ? CallHook : NULL;
+
+    InitBConOutPStream(&out, &bbs, con, R_pstream_xdr_format, 0, hook, fun);
+    R_Serialize(object, &out);
+    flush_bcon_buffer(&bbs);
+    return R_NilValue;
+}
+
+
+/*
+ * Persistent Memory Streams
+ */
+
+typedef struct membuf_st {
+    int size;
+    int count;
+    unsigned char *buf;
+} *membuf_t;
+
+static void resize_buffer(membuf_t mb, int needed)
+{
+    int newsize = 2 * needed;
+    mb->buf = realloc(mb->buf, newsize);
+    if (mb->buf == NULL)
+	error(_("cannot allocate buffer"));
+}
+
+static void OutCharMem(R_outpstream_t stream, int c)
+{
+    membuf_t mb = stream->data;
+    if (mb->count >= mb->size)
+	resize_buffer(mb, mb->count + 1);
+    mb->buf[mb->count++] = c;
+}
+
+static void OutBytesMem(R_outpstream_t stream, void *buf, int length)
+{
+    membuf_t mb = stream->data;
+    if (mb->count + length > mb->size)
+	resize_buffer(mb, mb->count + length);
+    memcpy(mb->buf + mb->count, buf, length);
+    mb->count += length;
+}
+
+static int InCHarMem(R_inpstream_t stream)
+{
+    membuf_t mb = stream->data;
+    if (mb->count >= mb->size)
+	error(_("read error"));
+    return mb->buf[mb->count++];
+}
+
+static void InBytesMem(R_inpstream_t stream, void *buf, int length)
+{
+    membuf_t mb = stream->data;
+    if (mb->count + length > mb->size)
+	error(_("read error"));
+    memcpy(buf, mb->buf + mb->count, length);
+    mb->count += length;
+}
+
+static void InitMemInPStream(R_inpstream_t stream, membuf_t mb,
+			     void *buf, int length,
+			     SEXP (*phook)(SEXP, SEXP), SEXP pdata)
+{
+    mb->count = 0;
+    mb->size = length;
+    mb->buf = buf;
+    R_InitInPStream(stream, (R_pstream_data_t) mb, R_pstream_any_format,
+		    InCHarMem, InBytesMem, phook, pdata);
+}
+
+static void InitMemOutPStream(R_outpstream_t stream, membuf_t mb,
+			      R_pstream_format_t type, int version,
+			      SEXP (*phook)(SEXP, SEXP), SEXP pdata)
+{
+    mb->count = 0;
+    mb->size = 0;
+    mb->buf = NULL;
+    R_InitOutPStream(stream, (R_pstream_data_t) mb, type, version,
+		     OutCharMem, OutBytesMem, phook, pdata);
+}
+
+static void free_mem_buffer(void *data)
+{
+    membuf_t mb = data;
+    if (mb->buf != NULL) {
+	unsigned char *buf = mb->buf;
+	mb->buf = NULL;
+	free(buf);
+    }
+}
+    
+static SEXP CloseMemOutPStream(R_outpstream_t stream)
+{
+    SEXP val;
+    membuf_t mb = stream->data;
+    PROTECT(val = allocVector(CHARSXP, mb->count));
+    memcpy(CHAR(val), mb->buf, mb->count);
+    val = ScalarString(val);
+    free_mem_buffer(mb);
+    UNPROTECT(1);
+    return val;
+}
+
+SEXP R_serialize(SEXP object, SEXP icon, SEXP ascii, SEXP fun)
+{
+    struct R_outpstream_st out;
+    R_pstream_format_t type;
+    SEXP (*hook)(SEXP, SEXP);
+
+    hook = fun != R_NilValue ? CallHook : NULL;
+
+    if (asLogical(ascii)) type = R_pstream_ascii_format;
+    else type = R_pstream_xdr_format; /**** binary or ascii if no XDR? */
+
+    if (icon == R_NilValue) {
+	RCNTXT cntxt;
+	struct membuf_st mbs;
+	SEXP val;
+
+	/* set up a context which will free the buffer if there is an error */
+	begincontext(&cntxt, CTXT_CCODE, R_NilValue, R_NilValue, R_NilValue,
+		     R_NilValue, R_NilValue);
+	cntxt.cend = &free_mem_buffer;
+	cntxt.cenddata = &mbs;
+
+	InitMemOutPStream(&out, &mbs, type, 0, hook, fun);
+	R_Serialize(object, &out);
+
+	val =  CloseMemOutPStream(&out);
+
+	/* end the context after anything that could raise an error but before
+	   calling OutTerm so it doesn't get called twice */
+	endcontext(&cntxt);
+
+	return val;
+    }
+    else {
+	Rconnection con = getConnection(asInteger(icon));
+	R_InitConnOutPStream(&out, con, type, 0, hook, fun);
+	R_Serialize(object, &out);
+	return R_NilValue;
+    }
+}
+
+SEXP R_unserialize(SEXP icon, SEXP fun)
+{
+    struct R_inpstream_st in;
+    SEXP (*hook)(SEXP, SEXP);
+
+    hook = fun != R_NilValue ? CallHook : NULL;
+
+    if (TYPEOF(icon) == STRSXP && LENGTH(icon) > 0) {
+	struct membuf_st mbs;
+	void *data = CHAR(STRING_ELT(icon, 0));
+	int length = LENGTH(STRING_ELT(icon, 0));
+	InitMemInPStream(&in, &mbs, data,  length, hook, fun);
+	return R_Unserialize(&in);
+    }
+    else {
+	Rconnection con = getConnection(asInteger(icon));
+	R_InitConnInPStream(&in, con, R_pstream_any_format, hook, fun);
+	return R_Unserialize(&in);
+    }
+}
+
+
+/*
+ * Support Code for Lazy Loading of Packages
+ */
+
+/* This include is to bring in declarations of R_compress1 and
+   R_decompress1 */
+#include "basedecl.h"
+
+#define IS_PROPER_STRING(s) (TYPEOF(s) == STRSXP && LENGTH(s) > 0)
+
+/* Appends a scalar string to the end of a file using binary mode.
+   Returns an integer vector of the initial offset of the string in
+   the file and the length of the string. */
+
+static SEXP appendStringToFile(SEXP file, SEXP string)
+{
+    FILE *fp;
+    size_t len, out;
+    long pos;
+    SEXP val;
+
+    if (! IS_PROPER_STRING(file))
+	error(_("not a proper file name"));
+    if (! IS_PROPER_STRING(string))
+	error(_("not a proper string"));
+#ifdef HAVE_WORKING_FTELL
+    /* Windows' ftell returns position 0 with "ab" */
+    if ((fp = fopen(CHAR(STRING_ELT(file, 0)), "ab")) == NULL)
+	error(_("file open failed"));
+#else
+    if ((fp = fopen(CHAR(STRING_ELT(file, 0)), "r+b")) == NULL)
+	 error(_("file open failed"));
+    fseek(fp, 0, SEEK_END);
+#endif
+
+    len = LENGTH(STRING_ELT(string, 0));
+    pos = ftell(fp);
+    out = fwrite(CHAR(STRING_ELT(string, 0)), 1, len, fp);
+    fclose(fp);
+
+    if (out != len) error(_("write failed"));
+    if (pos == -1) error(_("could not determine file position"));
+
+    val = allocVector(INTSXP, 2);
+    INTEGER(val)[0] = pos;
+    INTEGER(val)[1] = len;
+    return val;
+}
+
+/* Experimental interface to cache the pkg.rdb files */
+
+#define NC 100
+static int used = 0;
+static char names[NC][PATH_MAX];
+static char *ptr[NC];
+
+SEXP R_lazyLoadDBflush(SEXP file)
+{
+    int i;
+    char *cfile = CHAR(STRING_ELT(file, 0));
+
+    /* fprintf(stderr, "flushing file %s", cfile); */
+    for (i = 0; i < used; i++)
+	if(strcmp(cfile, names[i]) == 0) {
+	    strcpy(names[i], "");
+	    free(ptr[i]);
+	    /* fprintf(stderr, " found at pos %d in cache", i); */
+	    break;
+	}
+    /* fprintf(stderr, "\n"); */
+    return R_NilValue;
+}
+
+
+/* Reads, in binary mode, the bytes in the range specified by a
+   position/length vector and returns them as a scalar string. */
+
+static SEXP readStringFromFile(SEXP file, SEXP key)
+{
+    FILE *fp;
+    int offset, len, in, i, icache = -1, filelen;
+    SEXP val;
+    char *cfile = CHAR(STRING_ELT(file, 0));
+
+    if (! IS_PROPER_STRING(file))
+	error(_("not a proper file name"));
+    if (TYPEOF(key) != INTSXP || LENGTH(key) != 2)
+	error(_("bad offset/length argument"));
+
+    offset = INTEGER(key)[0];
+    len = INTEGER(key)[1];
+
+    val = allocVector(CHARSXP, len);
+    val = ScalarString(val);
+    /* Do we have this database cached? */
+    for (i = 0; i < used; i++)
+	if(strcmp(cfile, names[i]) == 0) {icache = i; break;}
+    if (icache >= 0) {
+	memcpy(CHAR(STRING_ELT(val, 0)), ptr[icache]+offset, len);
+	return val;
+    }
+
+    /* find a vacant slot? */
+    for (i = 0; i < used; i++)
+	if(strcmp("", names[i]) == 0) {icache = i; break;}
+    if(icache < 0 && used < NC) icache = used++;
+
+    if(icache >= 0) { 
+	strcpy(names[icache], cfile);
+	if ((fp = fopen(cfile, "rb")) == NULL)
+	    error(_("open failed on %s"), cfile);
+	if (fseek(fp, 0, SEEK_END) != 0) {
+	    fclose(fp);
+	    error(_("seek failed on %s"), cfile);
+	}
+	filelen = ftell(fp);
+	/* fprintf(stderr, "adding file %s at pos %d in cache, length %d\n", 
+	   cfile, icache, filelen); */
+	ptr[icache] = malloc(filelen);
+	if (fseek(fp, 0, SEEK_SET) != 0) {
+	    fclose(fp);
+	    error(_("seek failed on %s"), cfile);
+	}
+	in = fread(ptr[icache], 1, filelen, fp);
+	fclose(fp);
+	if (filelen != in) error(_("read failed on %s"), cfile);
+	memcpy(CHAR(STRING_ELT(val, 0)), ptr[icache]+offset, len);	
+    } else {
+	if ((fp = fopen(cfile, "rb")) == NULL)
+	    error(_("open failed on %s"), cfile);
+	if (fseek(fp, offset, SEEK_SET) != 0) {
+	    fclose(fp);
+	    error(_("seek failed on %s"), cfile);
+	}
+	in = fread(CHAR(STRING_ELT(val, 0)), 1, len, fp);
+	fclose(fp);
+	if (len != in) error(_("read failed on %s"), cfile);
+    }
+    
+    return val;
+}
+
+/* Gets the binding values of variables from a frame and returns them
+   as a list.  If the force argument is true, promises are forced;
+   otherwise they are not. */
+
+SEXP R_getVarsFromFrame(SEXP vars, SEXP env, SEXP forcesxp)
+{
+    SEXP val, tmp, sym;
+    Rboolean force;
+    int i, len;
+
+    if (TYPEOF(env) != NILSXP && TYPEOF(env) != ENVSXP)
+        error(_("bad environment"));
+    if (TYPEOF(vars) != STRSXP)
+        error(_("bad variable names"));
+    force = asLogical(forcesxp);
+
+    len = LENGTH(vars);
+    PROTECT(val = allocVector(VECSXP, len));
+    for (i = 0; i < len; i++) {
+	sym = install(CHAR(STRING_ELT(vars, i)));
+        if (TYPEOF(env) == NILSXP)
+	    tmp = findVar(sym, env);
+	else
+	    tmp = findVarInFrame(env, sym);
+	if (tmp == R_UnboundValue)
+	    error(_("object '%s' not found"), CHAR(STRING_ELT(vars, i)));
+        if (force && TYPEOF(tmp) == PROMSXP) {
+            PROTECT(tmp);
+            tmp = eval(tmp, R_GlobalEnv);
+            SET_NAMED(tmp, 2);
+            UNPROTECT(1);
+        }
+        else if (TYPEOF(tmp) != NILSXP && NAMED(tmp) < 1)
+            SET_NAMED(tmp, 1);
+	SET_VECTOR_ELT(val, i, tmp);
+    }
+    setAttrib(val, R_NamesSymbol, vars);    
+    UNPROTECT(1);
+
+    return val;
+}
+
+/* Serializes and, optionally, compresses a value and appends the
+   result to a file.  Returns the key position/length key for
+   retrieving the value */
+
+SEXP R_lazyLoadDBinsertValue(SEXP value, SEXP file, SEXP ascii,
+			     SEXP compsxp, SEXP hook)
+{
+    PROTECT_INDEX vpi;
+    Rboolean compress = asLogical(compsxp);
+    SEXP key;
+
+    value = R_serialize(value, R_NilValue, ascii, hook);
+    PROTECT_WITH_INDEX(value, &vpi);
+    if (compress)
+	REPROTECT(value = R_compress1(value), vpi);
+    key = appendStringToFile(file, value);
+    UNPROTECT(1);
+    return key;
+}
+
+/* Retrieves a sequence of bytes as specified by a position/length key
+   from a file, optionally decompresses, and unserializes the bytes.
+   If the result is a promise, then the promise is forced. */
+
+SEXP R_lazyLoadDBfetch(SEXP key, SEXP file, SEXP compsxp, SEXP hook)
+{
+    PROTECT_INDEX vpi;
+    Rboolean compressed = asLogical(compsxp);
+    SEXP val;
+
+    PROTECT_WITH_INDEX(val = readStringFromFile(file, key), &vpi);
+    if (compressed)
+	REPROTECT(val = R_decompress1(val), vpi);
+    val = R_unserialize(val, hook);
+    if (TYPEOF(val) == PROMSXP) {
+        REPROTECT(val, vpi);
+        val = eval(val, R_GlobalEnv);
+        SET_NAMED(val, 2);
+    }
+    UNPROTECT(1);
+    return val;
+}
