@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2000-3   The R Development Core Team.
+ *  Copyright (C) 2000-5   The R Development Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -17,6 +17,8 @@
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
+/* <UTF8> the only interpretation of char is ASCII */
+
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #endif
@@ -35,7 +37,7 @@ static int   in_R_FTPRead(void *ctx, char *dest, int len);
 static void  in_R_FTPClose(void *ctx);
 
 
-#include <R_ext/Rinternet.h>
+#include <Rmodules/Rinternet.h>
 
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
@@ -70,7 +72,7 @@ static Rboolean url_open(Rconnection con)
 	if(ctxt == NULL) {
 	  /* if we call error() we get a connection leak*/
 	  /* so do_url has to raise the error*/
-	  /* error("cannot open URL `%s'", url); */
+	  /* error("cannot open URL '%s'", url); */
 	    return FALSE;
 	}
 	((Rurlconn)(con->private))->ctxt = ctxt;
@@ -80,13 +82,13 @@ static Rboolean url_open(Rconnection con)
 	if(ctxt == NULL) {
 	  /* if we call error() we get a connection leak*/
 	  /* so do_url has to raise the error*/
-	  /* error("cannot open URL `%s'", url); */
+	  /* error("cannot open URL '%s'", url); */
 	    return FALSE;
 	}
 	((Rurlconn)(con->private))->ctxt = ctxt;
 	break;
     default:
-	warning("unknown URL scheme");
+	warning(_("unknown URL scheme"));
 	return FALSE;
     }
 
@@ -113,7 +115,7 @@ static void url_close(Rconnection con)
     con->isopen = FALSE;
 }
 
-static int url_fgetc(Rconnection con)
+static int url_fgetc_internal(Rconnection con)
 {
     UrlScheme type = ((Rurlconn)(con->private))->type;
     void * ctxt = ((Rurlconn)(con->private))->ctxt;
@@ -155,28 +157,29 @@ static Rconnection in_R_newurl(char *description, char *mode)
     Rconnection new;
 
     new = (Rconnection) malloc(sizeof(struct Rconn));
-    if(!new) error("allocation of url connection failed");
+    if(!new) error(_("allocation of url connection failed"));
     new->class = (char *) malloc(strlen("file") + 1);
     if(!new->class) {
 	free(new);
-	error("allocation of url connection failed");
+	error(_("allocation of url connection failed"));
     }
     strcpy(new->class, "url");
     new->description = (char *) malloc(strlen(description) + 1);
     if(!new->description) {
 	free(new->class); free(new);
-	error("allocation of url connection failed");
+	error(_("allocation of url connection failed"));
     }
     init_con(new, description, mode);
     new->canwrite = FALSE;
     new->open = &url_open;
     new->close = &url_close;
-    new->fgetc = &url_fgetc;
+    new->fgetc_internal = &url_fgetc_internal;
+    new->fgetc = &dummy_fgetc;
     new->read = &url_read;
     new->private = (void *) malloc(sizeof(struct urlconn));
     if(!new->private) {
 	free(new->description); free(new->class); free(new);
-	error("allocation of url connection failed");
+	error(_("allocation of url connection failed"));
     }
 
     IDquiet = TRUE;
@@ -197,6 +200,14 @@ static void putdots(int *pold, int new)
     }
     if(R_Consolefile) fflush(R_Consolefile);
 }
+
+static void putdashes(int *pold, int new)
+{
+    int i, old = *pold;
+    *pold = new;
+    for(i = old; i < new; i++)  REprintf("=");
+    if(R_Consolefile) fflush(R_Consolefile);
+}
 #endif
 
 /* note, ALL the possible structures have the first two elements */
@@ -229,26 +240,26 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
     checkArity(op, args);
     scmd = CAR(args); args = CDR(args);
     if(!isString(scmd) || length(scmd) < 1)
-	error("invalid `url' argument");
+	error(_("invalid 'url' argument"));
     if(length(scmd) > 1)
-	warning("only first element of `url' argument used");
+	warning(_("only first element of 'url' argument used"));
     url = CHAR(STRING_ELT(scmd, 0));
     sfile = CAR(args); args = CDR(args);
     if(!isString(sfile) || length(sfile) < 1)
-	error("invalid `destfile' argument");
+	error(_("invalid 'destfile' argument"));
     if(length(sfile) > 1)
-	warning("only first element of `destfile' argument used");
+	warning(_("only first element of 'destfile' argument used"));
     file = CHAR(STRING_ELT(sfile, 0));
     IDquiet = quiet = asLogical(CAR(args)); args = CDR(args);
     if(quiet == NA_LOGICAL)
-	error("invalid `quiet' argument");
+	error(_("invalid 'quiet' argument"));
     smode =  CAR(args); args = CDR(args);
     if(!isString(smode) || length(smode) != 1)
-	error("invalid `mode' argument");
+	error(_("invalid 'mode' argument"));
     mode = CHAR(STRING_ELT(smode, 0));
     cacheOK = asLogical(CAR(args));
     if(cacheOK == NA_LOGICAL)
-	error("invalid `cacheOK' argument");
+	error(_("invalid 'cacheOK' argument"));
 
     if(strncmp(url, "file://", 7) == 0) {
 	FILE *in, *out;
@@ -257,9 +268,9 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 
 	/* Use binary transfers */
 	in = R_fopen(R_ExpandFileName(url+7), (mode[2] == 'b') ? "rb" : "r");
-	if(!in) error("cannot open URL `%s'", url);
+	if(!in) error(_("cannot open URL '%s'"), url);
 	out = R_fopen(R_ExpandFileName(file), mode);
-	if(!out) error("cannot open destfile `%s'", file);
+	if(!out) error(_("cannot open destfile '%s'"), file);
 	while((n = fread(buf, 1, CPBUFSIZE, in)) > 0)
 	    fwrite(buf, 1, n, out);
 	fclose(out); fclose(in);
@@ -269,29 +280,29 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 
 	FILE *out;
 	void *ctxt;
-	int len, total, guess, nnew, nbytes = 0;
+	int len, total, guess, nbytes = 0;
 	char buf[IBUFSIZE];
 #ifndef Win32
 	int ndots = 0;
 #endif
 
 	out = R_fopen(R_ExpandFileName(file), mode);
-	if(!out) error("cannot open destfile `%s'", file);
+	if(!out) error(_("cannot open destfile '%s'"), file);
 
 	R_Busy(1);
-	if(!quiet) REprintf("trying URL `%s'\n", url);
+	if(!quiet) REprintf(_("trying URL '%s'\n"), url);
 #ifdef Win32
 	R_FlushConsole();
 #endif
 	ctxt = in_R_HTTPOpen(url, cacheOK);
 	if(ctxt == NULL) status = 1;
 	else {
-	    if(!quiet) REprintf("opened URL\n", url);
+	    if(!quiet) REprintf(_("opened URL\n"), url);
 	    guess = total = ((inetconn *)ctxt)->length;
-	    if (guess <= 0) guess = 100 * 1024;
 #ifdef Win32
+	    if (guess <= 0) guess = 100 * 1024;
 	    R_FlushConsole();
-	    wprog = newwindow("Download progress", rect(0, 0, 540, 100),
+	    wprog = newwindow(_("Download progress"), rect(0, 0, 540, 100),
 			      Titlebar | Centered);
 	    setbackground(wprog, dialog_bg());
 	    strcpy(buf, "URL: ");
@@ -306,7 +317,6 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 	    while ((len = in_R_HTTPRead(ctxt, buf, sizeof(buf))) > 0) {
 		fwrite(buf, 1, len, out);
 		nbytes += len;
-		nnew = nbytes/1024;
 #ifdef Win32
 		if(nbytes > guess) {
 		    guess *= 2;
@@ -314,7 +324,10 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 		}
 		setprogressbar(pb, nbytes);
 #else
-		if(!quiet) putdots(&ndots, nnew);
+		if(!quiet) {
+		    if(guess <= 0) putdots(&ndots, nbytes/1024);
+		    else putdashes(&ndots, 50*nbytes/guess);
+		}
 #endif
 	    }
 	    in_R_HTTPClose(ctxt);
@@ -336,39 +349,39 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 	    del(wprog);
 #endif
 	    if (total > 0 && total != nbytes)
-		warning("downloaded length %d != reported length %d",
+		warning(_("downloaded length %d != reported length %d"),
 			nbytes, total);
 	}
 	R_Busy(0);
-	if (status == 1) error("cannot open URL `%s'", url);
+	if (status == 1) error(_("cannot open URL '%s'"), url);
 
     } else if (strncmp(url, "ftp://", 6) == 0) {
 
 	FILE *out;
 	void *ctxt;
-	int len, total, guess, nnew, nbytes = 0;
+	int len, total, guess, nbytes = 0;
 	char buf[IBUFSIZE];
 #ifndef Win32
 	int ndots = 0;
 #endif
 
 	out = R_fopen(R_ExpandFileName(file), mode);
-	if(!out) error("cannot open destfile `%s'", file);
+	if(!out) error(_("cannot open destfile '%s'"), file);
 
 	R_Busy(1);
-	if(!quiet) REprintf("trying URL `%s'\n", url);
+	if(!quiet) REprintf(_("trying URL '%s'\n"), url);
 #ifdef Win32
 	R_FlushConsole();
 #endif
 	ctxt = in_R_FTPOpen(url);
 	if(ctxt == NULL) status = 1;
 	else {
-	    if(!quiet) REprintf("opened URL\n", url);
+	    if(!quiet) REprintf(_("opened URL\n"), url);
 	    guess = total = ((inetconn *)ctxt)->length;
-	    if (guess <= 0) guess = 100 * 1024;
 #ifdef Win32
+	    if (guess <= 0) guess = 100 * 1024;
 	    R_FlushConsole();
-	    wprog = newwindow("Download progress", rect(0, 0, 540, 100),
+	    wprog = newwindow(_("Download progress"), rect(0, 0, 540, 100),
 			      Titlebar | Centered);
 	    setbackground(wprog, LightGrey);
 	    strcpy(buf, "URL: ");
@@ -383,7 +396,6 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 	    while ((len = in_R_FTPRead(ctxt, buf, sizeof(buf))) > 0) {
 		fwrite(buf, 1, len, out);
 		nbytes += len;
-		nnew = nbytes/1024;
 #ifdef Win32
 		if(nbytes > guess) {
 		    guess *= 2;
@@ -391,7 +403,10 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 		}
 		setprogressbar(pb, nbytes);
 #else
-		if(!quiet) putdots(&ndots, nnew);
+		if(!quiet) {
+		    if(guess <= 0) putdots(&ndots, nbytes/1024);
+		    else putdashes(&ndots, 50*nbytes/guess);
+		}
 #endif
 	    }
 	    in_R_FTPClose(ctxt);
@@ -413,15 +428,15 @@ static SEXP in_do_download(SEXP call, SEXP op, SEXP args, SEXP env)
 	    del(wprog);
 #endif
 	    if (total > 0 && total != nbytes)
-		warning("downloaded length %d != reported length %d",
+		warning(_("downloaded length %d != reported length %d"),
 			nbytes, total);
 	}
 	R_Busy(0);
-	if (status == 1) error("cannot open URL `%s'", url);
+	if (status == 1) error(_("cannot open URL '%s'"), url);
 #endif
 
     } else
-	error("unsupported URL scheme");
+	error(_("unsupported URL scheme"));
 
     PROTECT(ans = allocVector(INTSXP, 1));
     INTEGER(ans)[0] = status;
@@ -448,14 +463,14 @@ void *in_R_HTTPOpen(const char *url, int cacheOK)
 	int rc = RxmlNanoHTTPReturnCode(ctxt);
 	if(rc != 200) {
 	    RxmlNanoHTTPClose(ctxt);
-	    warning("cannot open: HTTP status was `%d %s'", rc,
+	    warning(_("cannot open: HTTP status was '%d %s'"), rc,
 		  RxmlNanoHTTPStatusMsg(ctxt));
 	    return NULL;
 	} else {
 	    type = RxmlNanoHTTPContentType(ctxt);
 	    len = RxmlNanoHTTPContentLength(ctxt);
 	    if(!IDquiet){
-		REprintf("Content type `%s'", type ? type : "unknown");
+		REprintf("Content type '%s'", type ? type : "unknown");
 		if(len >= 0) REprintf(" length %d bytes\n", len);
 		else REprintf(" length unknown\n", len);
 #ifdef Win32
@@ -618,7 +633,7 @@ static void *in_R_HTTPOpen(const char *url, const int cacheOK)
 	if(callback_status != INTERNET_STATUS_REQUEST_COMPLETE) {
 	    InternetCloseHandle(wictxt->hand);
 	    free(wictxt);
-	    warning("InternetOpenUrl timed out");
+	    warning(_("InternetOpenUrl timed out"));
 	    return NULL;
 	}
     }
@@ -645,7 +660,7 @@ static void *in_R_HTTPOpen(const char *url, const int cacheOK)
 		p = buf + strlen(buf) - 1;
 		if(*p == '\n' || *p == '\r') *p = '\0'; else break;
 	    }
-	    warning("InternetOpenUrl failed: `%s'", buf);
+	    warning(_("InternetOpenUrl failed: '%s'"), buf);
 	    return NULL;
 	} else {
 	    FormatMessage( 
@@ -659,7 +674,7 @@ static void *in_R_HTTPOpen(const char *url, const int cacheOK)
 		p = buf + strlen(buf) - 1;
 		if(*p == '\n' || *p == '\r') *p = '\0'; else break;
 	    }
-	    warning("InternetOpenUrl failed: `%s'", buf);
+	    warning(_("InternetOpenUrl failed: '%s'"), buf);
 	    return NULL;
 	}
     }
@@ -674,7 +689,7 @@ static void *in_R_HTTPOpen(const char *url, const int cacheOK)
 	InternetCloseHandle(wictxt->session);
 	InternetCloseHandle(wictxt->hand);
 	free(wictxt);
-	warning("cannot open: HTTP status was `%d %s'", status, buf);
+	warning(_("cannot open: HTTP status was '%d %s'"), status, buf);
 	return NULL;
     }
 
@@ -687,7 +702,7 @@ static void *in_R_HTTPOpen(const char *url, const int cacheOK)
     wictxt->length = status;
     wictxt->type = strdup(buf);
     if(!IDquiet) {
-	REprintf("Content type `%s' length %d bytes\n", buf, status);
+	REprintf("Content type '%s' length %d bytes\n", buf, status);
 	R_FlushConsole();
     }
 
@@ -709,7 +724,7 @@ static int in_R_HTTPRead(void *ctx, char *dest, int len)
 	    Sleep(100);
 	}
 	if(callback_status != INTERNET_STATUS_REQUEST_COMPLETE) {
-	    warning("Internet read timed out");
+	    warning(_("Internet read timed out"));
 	    nread = 0;
 	}
     }
@@ -774,7 +789,7 @@ static void *in_R_FTPOpen(const char *url)
 	if(callback_status != INTERNET_STATUS_REQUEST_COMPLETE) {
 	    InternetCloseHandle(wictxt->hand);
 	    free(wictxt);
-	    warning("InternetOpenUrl timed out");
+	    warning(_("InternetOpenUrl timed out"));
 	    return NULL;
 	}
     }
@@ -796,7 +811,7 @@ static void *in_R_FTPOpen(const char *url)
 	free(wictxt);
 	if (err1 == ERROR_INTERNET_EXTENDED_ERROR) {
 	    InternetGetLastResponseInfo(&err2, buf, &blen);
-	    warning("InternetOpenUrl failed: `%s'", buf);
+	    warning(_("InternetOpenUrl failed: '%s'"), buf);
 	    return NULL;
 	} else {
 	    FormatMessage( 
@@ -805,7 +820,7 @@ static void *in_R_FTPOpen(const char *url)
 		err1,
 		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
 		buf, 101, NULL);
-	    warning("InternetOpenUrl failed: `%s'", buf);
+	    warning(_("InternetOpenUrl failed: '%s'"), buf);
 	    return NULL;
 	}
     }
@@ -888,7 +903,7 @@ void R_init_internet(DllInfo *info)
     R_InternetRoutines *tmp;
     tmp = (R_InternetRoutines*) malloc(sizeof(R_InternetRoutines));
     if(!tmp) {
-	error("Cannot allocate memory for InternetRoutines structure");
+	error(_("cannot allocate memory for InternetRoutines structure"));
 	return;
     }
 
