@@ -1,6 +1,6 @@
 /*
  *  Mathlib : A C Library of Special Functions
- *  Copyright (C) 1998-2003  The R Development Core Team
+ *  Copyright (C) 1998-2004  The R Development Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +30,9 @@
 
 #ifndef MATHLIB_STANDALONE
 /* Mathlib in R */
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
 
 #include <R_ext/Error.h>
 # define MATHLIB_ERROR(fmt,x)		error(fmt,x);
@@ -46,47 +49,38 @@
 void R_CheckUserInterrupt(void);
 #define calloc R_chk_calloc
 #define free R_chk_free
+
+#ifdef ENABLE_NLS
+#include <libintl.h>
+#define _(String) gettext (String)
+#else
+#define _(String) (String)
+#endif
+
 #else
 /* Mathlib standalone */
 
 #include <stdio.h>
-# define MATHLIB_ERROR(fmt,x)	{ printf(fmt,x); exit(1); }
-# define MATHLIB_WARNING(fmt,x)		printf(fmt,x)
-# define MATHLIB_WARNING2(fmt,x,x2)	printf(fmt,x,x2)
-# define MATHLIB_WARNING3(fmt,x,x2,x3)	printf(fmt,x,x2,x3)
-# define MATHLIB_WARNING4(fmt,x,x2,x3,x4) printf(fmt,x,x2,x3,x4)
+#define MATHLIB_ERROR(fmt,x)	{ printf(fmt,x); exit(1); }
+#define MATHLIB_WARNING(fmt,x)		printf(fmt,x)
+#define MATHLIB_WARNING2(fmt,x,x2)	printf(fmt,x,x2)
+#define MATHLIB_WARNING3(fmt,x,x2,x3)	printf(fmt,x,x2,x3)
+#define MATHLIB_WARNING4(fmt,x,x2,x3,x4) printf(fmt,x,x2,x3,x4)
 
-#ifdef IEEE_754
-# define ISNAN(x) (isnan(x)!=0)
-#else
-# define ISNAN(x)      R_IsNaNorNA(x)
-#endif
+#define ISNAN(x) (isnan(x)!=0)
 #define R_FINITE(x)    R_finite(x)
-int R_IsNaNorNA(double);
 int R_finite(double);
 
-#ifdef IEEE_754
 #define ML_POSINF	(1.0 / 0.0)
 #define ML_NEGINF	((-1.0) / 0.0)
 #define ML_NAN		(0.0 / 0.0)
-#else
-#define ML_POSINF	DBL_MAX
-#define ML_NEGINF	(-DBL_MAX)
-#define ML_NAN		(-DBL_MAX*(1-1e-15))
-#endif
 
+#define _(String) String
 #endif /* standalone */
 
-#ifdef IEEE_754
 #define ML_ERROR(x)	/* nothing */
 #define ML_UNDERFLOW	(DBL_MIN * DBL_MIN)
 #define ML_VALID(x)	(!ISNAN(x))
-#else/*--- NO IEEE: No +/-Inf, NAN,... ---*/
-void ml_error(int n);
-#define ML_ERROR(x)	ml_error(x)
-#define ML_UNDERFLOW	0
-#define ML_VALID(x)	(errno == 0)
-#endif
 
 #define ME_NONE		0
 /*	no error */
@@ -117,7 +111,6 @@ void ml_error(int n);
 #define bd0       	Rf_bd0
 #define chebyshev_eval	Rf_chebyshev_eval
 #define chebyshev_init	Rf_chebyshev_init
-#define fastchoose	Rf_fastchoose
 #define i1mach		Rf_i1mach
 #define gammalims	Rf_gammalims
 #define lfastchoose	Rf_lfastchoose
@@ -135,7 +128,6 @@ void	gammalims(double*, double*);
 double	lgammacor(double); /* log(gamma) correction */
 double  stirlerr(double);  /* Stirling expansion "error" */
 
-double	fastchoose(double, double);
 double	lfastchoose(double, double);
 
 double  bd0(double, double);
@@ -147,5 +139,9 @@ double  pnchisq_raw(double, double, double, double, double, int);
 
 int	i1mach(int);
 
+/* From toms708.c */
+void bratio(double a, double b, double x, double y, 
+	    double *w, double *w1, int *ierr);
+
 
 #endif /* MATHLIB_PRIVATE_H */
