@@ -1,6 +1,6 @@
 /*
  *  Mathlib : A C Library of Special Functions
- *  Copyright (C) 2000  The R Development Core Team
+ *  Copyright (C) 2000, 2003  The R Development Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -27,6 +27,11 @@ void set_seed(unsigned int i1, unsigned int i2)
     I1 = i1; I2 = i2;
 }
 
+void get_seed(unsigned int *i1, unsigned int *i2)
+{
+    *i1 = I1; *i2 = I2;
+}
+
 
 double unif_rand(void)
 {
