@@ -4,6 +4,8 @@ postscript("reg-plot.ps", paper="a4", horizontal=TRUE,
            encoding ="ISOLatin1.enc")
 ## since we supply the font metrics, the results depend only on
 ## the encoding used: Windows is different from Unix by default.
+## As from 2.1.0 we only test genuine Latin-1 chars: Adobe has dotlessi
+## and accents at 144:152, 154, 155, 157:159, but Unicode does not.
 
 options(warn = 1)# print them as they occur
 
@@ -20,7 +22,7 @@ par(pty="s")
 plot(c(-1,16), c(-1,16), type="n", xlab="", ylab="", xaxs="i", yaxs="i")
 title("Centred chars in default char set (ISO Latin1)")
 grid(17, 17, lty=1)
-known <- c(32:126, 144:152, 154, 155, 157:255)
+known <- c(32:126, 160:255)
 
 for(i in known) {
     x <- i %% 16
@@ -34,7 +36,6 @@ par(pty="m")
 
 ### Prior to 1.2.2, the label sizes were unaffected by cex.
 
-data(VADeaths)
 dotchart(VADeaths, main = "Death Rates in Virginia - 1940", cex = 0.5)
 dotchart(VADeaths, main = "Death Rates in Virginia - 1940", cex = 1.5)
 
@@ -52,7 +53,6 @@ plot(1:10, type="n")
 polygon(c(1, 3, 3, 1), c(1, 1, 3, 3), col="yellow", border="red", lty="blank")
 rect(6, 6, 10, 10,  col="blue", border="red", lty="blank")
 ## in 1.5.0 all omit the fill colours.
-data(trees)
 with(trees, symbols(Height, Volume, circles=Girth/24, inches=FALSE,
                     lty="blank", bg="blue"))
 ## in 1.5.0 ignored the lty.
@@ -71,7 +71,6 @@ axis(1, pos = 0, at=c(-2,2,3), lty = lt, col=4)
 mtext(side=1,"note the x-ticks on the other side of the bars")
 
 ## plot.table(): explicit xlab and ylab for non-1D
-data(UCBAdmissions)
 plot(UCBAdmissions)# default x- and y-lab
 plot(UCBAdmissions, xlab = "x label", ylab = "YY")# wrong in 1.5.1
 ##   axis suppression
@@ -90,3 +89,12 @@ plot(ecdf(c(1:4,8,12)), ylab = "ECDF", main=NULL)
 plot(1:10, pch = NA) # gave error till 1.9.0
 points(1:3, pch=c("o",NA,"x"))# used "N"
 try(points(4, pch=c(NA,FALSE)))# still give an error
+
+## 'lwd' should transfer to plot symbols
+legend(1,10, c("A","bcd"), lwd = 2:3, pch= 21:22, pt.bg="skyblue",
+       col = 2:3, bg = "thistle")
+## (gave an error for 2 days in "2.0.0 unstable")
+
+x <- 2^seq(1,1001, length=20)
+plot(x, x^0.9, type="l", log="xy")
+## gave error 'Infinite axis extents [GEPretty(1.87013e-12,inf,5)]' for R 2.0.1
